"""Library of default cleaning functions."""

from matchbox.client.clean.lib import (
    company_name,
    company_number,
    drop,
    extract_cdms_number_to_new,
    extract_company_number_to_new,
    extract_duns_number_to_new,
    postcode,
    postcode_to_area,
)
from matchbox.client.clean.utils import alias, cleaning_function, unnest_renest

__all__ = (
    # Cleaning functions
    "company_name",
    "company_number",
    "drop",
    "extract_cdms_number_to_new",
    "extract_company_number_to_new",
    "extract_duns_number_to_new",
    "postcode",
    "postcode_to_area",
    # Utility functions
    "alias",
    "cleaning_function",
    "unnest_renest",
)



"""Functions abstracting the interaction with the server API."""

import time
from collections.abc import Iterable
from importlib.metadata import version
from io import BytesIO

import httpx
from pyarrow import Table
from pyarrow.parquet import read_table

from matchbox.client._settings import ClientSettings, settings
from matchbox.common.arrow import SCHEMA_MB_IDS, table_to_buffer
from matchbox.common.dtos import (
    BackendRetrievableType,
    ModelAncestor,
    ModelMetadata,
    ModelOperationStatus,
    NotFoundError,
    UploadStatus,
)
from matchbox.common.exceptions import (
    MatchboxClientFileError,
    MatchboxDeletionNotConfirmed,
    MatchboxResolutionNotFoundError,
    MatchboxServerFileError,
    MatchboxSourceNotFoundError,
    MatchboxUnhandledServerResponse,
    MatchboxUnparsedClientRequest,
)
from matchbox.common.graph import ResolutionGraph
from matchbox.common.hash import hash_to_base64
from matchbox.common.logging import logger
from matchbox.common.sources import Match, Source, SourceAddress

URLEncodeHandledType = str | int | float | bytes


def encode_param_value(
    v: URLEncodeHandledType | Iterable[URLEncodeHandledType],
) -> str | list[str]:
    if isinstance(v, str):
        return v
    elif isinstance(v, (int, float)):
        return str(v)
    elif isinstance(v, bytes):
        return hash_to_base64(v)
    # Needs to be at the end, so we don't apply it to e.g. strings
    if isinstance(v, Iterable):
        return [encode_param_value(item) for item in v]
    raise ValueError(f"It was not possible to parse {v} as an URL parameter")


def url_params(
    params: dict[str, URLEncodeHandledType | Iterable[URLEncodeHandledType]],
) -> dict[str, str | list[str]]:
    """Prepares a dictionary of parameters to be encoded in a URL."""
    non_null = {k: v for k, v in params.items() if v}
    return {k: encode_param_value(v) for k, v in non_null.items()}


def handle_http_code(res: httpx.Response) -> httpx.Response:
    """Handle HTTP status codes and raise appropriate exceptions."""
    res.read()

    if 299 >= res.status_code >= 200:
        return res

    if res.status_code == 400:
        if UploadStatus.model_validate_json(res.content, strict=False):
            error = UploadStatus.model_validate(res.json())
            raise MatchboxServerFileError(error.details)
        else:
            raise RuntimeError(f"Unexpected 400 error: {res.content}")

    if res.status_code == 404:
        error = NotFoundError.model_validate(res.json())
        if error.entity == BackendRetrievableType.SOURCE:
            raise MatchboxSourceNotFoundError(error.details)
        if error.entity == BackendRetrievableType.RESOLUTION:
            raise MatchboxResolutionNotFoundError(error.details)
        else:
            raise RuntimeError(f"Unexpected 404 error: {error.details}")

    if res.status_code == 409:
        error = ModelOperationStatus.model_validate(res.json())
        raise MatchboxDeletionNotConfirmed(message=error.details)

    if res.status_code == 422:
        raise MatchboxUnparsedClientRequest(res.content)

    raise MatchboxUnhandledServerResponse(res.content)


def create_client(settings: ClientSettings) -> httpx.Client:
    """Create an HTTPX client with proper configuration."""
    return httpx.Client(
        base_url=settings.api_root,
        timeout=httpx.Timeout(60 * 30, connect=settings.timeout, pool=settings.timeout),
        event_hooks={"response": [handle_http_code]},
        headers=create_headers(settings),
    )


def create_headers(settings: ClientSettings) -> dict[str, str]:
    """Creates client headers."""
    headers = {"X-Matchbox-Client-Version": version("matchbox_db")}
    if settings.api_key is not None:
        headers["X-API-Key"] = settings.api_key.get_secret_value()
    return headers


CLIENT = create_client(settings=settings)


# Retrieval


def query(
    source_address: SourceAddress,
    resolution_name: str | None = None,
    threshold: int | None = None,
    limit: int | None = None,
) -> Table:
    log_prefix = f"Query {source_address.pretty}"
    logger.debug(f"Using {resolution_name}", prefix=log_prefix)

    res = CLIENT.get(
        "/query",
        params=url_params(
            {
                "full_name": source_address.full_name,
                # Converted to b64 by `url_params()`
                "warehouse_hash_b64": source_address.warehouse_hash,
                "resolution_name": resolution_name,
                "threshold": threshold,
                "limit": limit,
            }
        ),
    )

    buffer = BytesIO(res.content)
    table = read_table(buffer)

    logger.debug("Finished", prefix=log_prefix)

    if not table.schema.equals(SCHEMA_MB_IDS):
        raise MatchboxClientFileError(
            message=(
                f"Schema mismatch. Expected:\n{SCHEMA_MB_IDS}\nGot:\n{table.schema}"
            )
        )

    return table


def match(
    targets: list[SourceAddress],
    source: SourceAddress,
    source_pk: str,
    resolution_name: str,
    threshold: int | None = None,
) -> Match:
    target_full_names = [t.full_name for t in targets]
    target_warehouse_hashes = [t.warehouse_hash for t in targets]

    log_prefix = f"Query {source.pretty}"
    logger.debug(
        f"{source_pk} to {', '.join(str(t) for t in targets)} using {resolution_name}",
        prefix=log_prefix,
    )

    res = CLIENT.get(
        "/match",
        params=url_params(
            {
                "target_full_names": target_full_names,
                # Converted to b64 by `url_params()`
                "target_warehouse_hashes_b64": target_warehouse_hashes,
                "source_full_name": source.full_name,
                # Converted to b64 by `url_params()`
                "source_warehouse_hash_b64": source.warehouse_hash,
                "source_pk": source_pk,
                "resolution_name": resolution_name,
                "threshold": threshold,
            }
        ),
    )

    logger.debug("Finished", prefix=log_prefix)

    return [Match.model_validate(m) for m in res.json()]


# Data management


def index(source: Source, batch_size: int | None = None) -> UploadStatus:
    """Index a Source in Matchbox."""
    log_prefix = f"Index {source.address.pretty}"
    log_batch = f"with batch size {batch_size:,}" if batch_size else "without batching"
    logger.debug(f"Started {log_batch}", prefix=log_prefix)

    logger.debug("Retrieving and hashing", prefix=log_prefix)

    data_hashes = source.hash_data(batch_size=batch_size)

    buffer = table_to_buffer(table=data_hashes)

    # Upload metadata
    logger.debug("Uploading metadata", prefix=log_prefix)

    metadata_res = CLIENT.post("/sources", json=source.model_dump())

    upload = UploadStatus.model_validate(metadata_res.json())

    # Upload data
    logger.debug("Uploading data", prefix=log_prefix)

    upload_res = CLIENT.post(
        f"/upload/{upload.id}",
        files={"file": (f"{upload.id}.parquet", buffer, "application/octet-stream")},
    )

    # Poll until complete with retry/timeout configuration
    status = UploadStatus.model_validate(upload_res.json())
    while status.status not in ["complete", "failed"]:
        status_res = CLIENT.get(f"/upload/{upload.id}/status")
        status = UploadStatus.model_validate(status_res.json())

        logger.debug(f"Uploading data: {status.status}", prefix=log_prefix)

        if status.status == "failed":
            raise MatchboxServerFileError(status.details)

        time.sleep(settings.retry_delay)

    logger.debug("Finished")

    return status


def get_source(address: SourceAddress) -> Source:
    log_prefix = f"Source {address.pretty}"
    logger.debug("Retrieving", prefix=log_prefix)

    res = CLIENT.get(f"/sources/{address.warehouse_hash_b64}/{address.full_name}")

    return Source.model_validate(res.json())


def get_resolution_sources(resolution_name: str) -> list[Source]:
    log_prefix = f"Resolution {resolution_name}"
    logger.debug("Retrieving", prefix=log_prefix)

    res = CLIENT.get("/sources", params={"resolution_name": resolution_name})

    return [Source.model_validate(s) for s in res.json()]


def get_resolution_graph() -> ResolutionGraph:
    """Get the resolution graph from Matchbox."""
    log_prefix = "Visualisation"
    logger.debug("Fetching resolution graph", prefix=log_prefix)

    res = CLIENT.get("/report/resolutions")
    return ResolutionGraph.model_validate(res.json())


# Model management


def insert_model(model: ModelMetadata) -> ModelOperationStatus:
    """Insert a model in Matchbox."""
    log_prefix = f"Model {model.name}"
    logger.debug("Inserting metadata", prefix=log_prefix)

    res = CLIENT.post("/models", json=model.model_dump())
    return ModelOperationStatus.model_validate(res.json())


def get_model(name: str) -> ModelMetadata:
    """Get model metadata from Matchbox."""
    log_prefix = f"Model {name}"
    logger.debug("Retrieving metadata", prefix=log_prefix)

    res = CLIENT.get(f"/models/{name}")
    return ModelMetadata.model_validate(res.json())


def add_model_results(name: str, results: Table) -> UploadStatus:
    """Upload model results in Matchbox."""
    log_prefix = f"Model {name}"
    logger.debug("Uploading results", prefix=log_prefix)

    buffer = table_to_buffer(table=results)

    # Initialise upload
    metadata_res = CLIENT.post(f"/models/{name}/results")

    upload = UploadStatus.model_validate(metadata_res.json())

    # Upload data
    upload_res = CLIENT.post(
        f"/upload/{upload.id}",
        files={"file": (f"{upload.id}.parquet", buffer, "application/octet-stream")},
    )

    logger.debug("Uploading data", prefix=log_prefix)

    # Poll until complete with retry/timeout configuration
    status = UploadStatus.model_validate(upload_res.json())
    while status.status not in ["complete", "failed"]:
        status_res = CLIENT.get(f"/upload/{upload.id}/status")
        status = UploadStatus.model_validate(status_res.json())

        logger.debug(f"Uploading data: {status.status}", prefix=log_prefix)

        if status.status == "failed":
            raise MatchboxServerFileError(status.details)

        time.sleep(settings.retry_delay)

    logger.debug("Finished", prefix=log_prefix)

    return status


def get_model_results(name: str) -> Table:
    """Get model results from Matchbox."""
    log_prefix = f"Model {name}"
    logger.debug("Retrieving results", prefix=log_prefix)

    res = CLIENT.get(f"/models/{name}/results")
    buffer = BytesIO(res.content)
    return read_table(buffer)


def set_model_truth(name: str, truth: int) -> ModelOperationStatus:
    """Set the truth threshold for a model in Matchbox."""
    log_prefix = f"Model {name}"
    logger.debug("Setting truth value", prefix=log_prefix)

    res = CLIENT.patch(f"/models/{name}/truth", json=truth)
    return ModelOperationStatus.model_validate(res.json())


def get_model_truth(name: str) -> int:
    """Get the truth threshold for a model in Matchbox."""
    log_prefix = f"Model {name}"
    logger.debug("Retrieving truth value", prefix=log_prefix)

    res = CLIENT.get(f"/models/{name}/truth")
    return res.json()


def get_model_ancestors(name: str) -> list[ModelAncestor]:
    """Get the ancestors of a model in Matchbox."""
    log_prefix = f"Model {name}"
    logger.debug("Retrieving ancestors", prefix=log_prefix)

    res = CLIENT.get(f"/models/{name}/ancestors")
    return [ModelAncestor.model_validate(m) for m in res.json()]


def set_model_ancestors_cache(
    name: str, ancestors: list[ModelAncestor]
) -> ModelOperationStatus:
    """Set the ancestors cache for a model in Matchbox."""
    log_prefix = f"Model {name}"
    logger.debug("Setting ancestors cached truth values", prefix=log_prefix)

    res = CLIENT.post(
        f"/models/{name}/ancestors_cache",
        json=[a.model_dump() for a in ancestors],
    )
    return ModelOperationStatus.model_validate(res.json())


def get_model_ancestors_cache(name: str) -> list[ModelAncestor]:
    """Get the ancestors cache for a model in Matchbox."""
    log_prefix = f"Model {name}"
    logger.debug("Getting ancestors cached truth values", prefix=log_prefix)

    res = CLIENT.get(f"/models/{name}/ancestors_cache")
    return [ModelAncestor.model_validate(m) for m in res.json()]


def delete_model(name: str, certain: bool = False) -> ModelOperationStatus:
    """Delete a model in Matchbox."""
    log_prefix = f"Model {name}"
    logger.debug("Deleting", prefix=log_prefix)

    res = CLIENT.delete(f"/models/{name}", params={"certain": certain})
    return ModelOperationStatus.model_validate(res.json())



"""Utilities for querying model results from the PostgreSQL backend."""

from typing import NamedTuple

from pyarrow import Table
from sqlalchemy import Engine, and_, case, func, select
from sqlalchemy.orm import Session

from matchbox.common.db import sql_to_df
from matchbox.common.dtos import ModelMetadata, ModelType
from matchbox.common.graph import ResolutionNodeType
from matchbox.server.postgresql.db import MBDB
from matchbox.server.postgresql.orm import (
    Contains,
    Probabilities,
    ResolutionFrom,
    Resolutions,
)
from matchbox.server.postgresql.utils.db import compile_sql


class SourceInfo(NamedTuple):
    """Information about a model's sources."""

    left: int
    right: int | None
    left_ancestors: set[int]
    right_ancestors: set[int] | None


def _get_model_parents(
    engine: Engine, resolution_id: int
) -> tuple[bytes, bytes | None]:
    """Get the model's immediate parent models."""
    parent_query = (
        select(Resolutions.resolution_id, Resolutions.type)
        .join(ResolutionFrom, Resolutions.resolution_id == ResolutionFrom.parent)
        .where(ResolutionFrom.child == resolution_id)
        .where(ResolutionFrom.level == 1)
    )

    with engine.connect() as conn:
        parents = conn.execute(parent_query).fetchall()

    if len(parents) == 1:
        return parents[0][0], None
    elif len(parents) == 2:
        p1, p2 = parents
        p1_id, p1_type = p1
        p2_id, p2_type = p2
        # Put dataset first if it exists
        if p1_type == ResolutionNodeType.DATASET:
            return p1_id, p2_id
        elif p2_type == ResolutionNodeType.DATASET:
            return p2_id, p1_id
        # Both models, maintain original order
        return p1_id, p2_id
    else:
        raise ValueError(f"Model has unexpected number of parents: {len(parents)}")


def _get_source_info(engine: Engine, resolution_id: int) -> SourceInfo:
    """Get source resolutions and their ancestry information."""
    left_id, right_id = _get_model_parents(engine=engine, resolution_id=resolution_id)

    with Session(engine) as session:
        left = session.get(Resolutions, left_id)
        right = session.get(Resolutions, right_id) if right_id else None

        left_ancestors = {left_id} | {m.resolution_id for m in left.ancestors}
        if right:
            right_ancestors = {right_id} | {m.resolution_id for m in right.ancestors}
        else:
            right_ancestors = None

    return SourceInfo(
        left=left_id,
        right=right_id,
        left_ancestors=left_ancestors,
        right_ancestors=right_ancestors,
    )


def get_model_metadata(engine: Engine, resolution: Resolutions) -> ModelMetadata:
    """Get metadata for a model resolution."""
    if resolution.type != ResolutionNodeType.MODEL:
        raise ValueError("Expected resolution of type model")

    source_info: SourceInfo = _get_source_info(
        engine=engine, resolution_id=resolution.resolution_id
    )

    with Session(engine) as session:
        left = session.get(Resolutions, source_info.left)
        right = (
            session.get(Resolutions, source_info.right) if source_info.right else None
        )

        return ModelMetadata(
            name=resolution.name,
            description=resolution.description or "",
            type=ModelType.DEDUPER if source_info.right is None else ModelType.LINKER,
            left_resolution=left.name,
            right_resolution=right.name if source_info.right else None,
        )


def get_model_results(resolution: Resolutions) -> Table:
    """Recover the model's pairwise probabilities and return as a PyArrow table.

    For each probability this model assigned:
    - Get its two immediate children
    - Filter for children that aren't parents of other clusters this model scored
    - Determine left/right by tracing ancestry to source resolutions using query helpers

    Args:
        resolution: Resolution of type model to query

    Returns:
        Table containing the original pairwise probabilities
    """
    engine = MBDB.get_engine()
    if resolution.type != ResolutionNodeType.MODEL:
        raise ValueError("Expected resolution of type model")

    source_info: SourceInfo = _get_source_info(
        engine=engine, resolution_id=resolution.resolution_id
    )

    # First get all clusters this resolution assigned probabilities to
    resolution_clusters = (
        select(Probabilities.cluster)
        .where(Probabilities.resolution == resolution.resolution_id)
        .cte("resolution_clusters")
    )

    # Get clusters that are parents in Contains for resolution's probabilities
    resolution_parents = (
        select(Contains.parent)
        .join(resolution_clusters, Contains.child == resolution_clusters.c.cluster)
        .cte("resolution_parents")
    )

    # Get valid pairs (those with exactly 2 children)
    # where neither child is a parent in the resolution's hierarchy
    valid_pairs = (
        select(Contains.parent)
        .join(
            Probabilities,
            and_(
                Probabilities.cluster == Contains.parent,
                Probabilities.resolution == resolution.resolution_id,
            ),
        )
        .where(~Contains.child.in_(select(resolution_parents)))
        .group_by(Contains.parent)
        .having(func.count() == 2)
        .cte("valid_pairs")
    )

    # Join to get children and probabilities
    pairs = (
        select(
            Contains.parent.label("id"),
            func.array_agg(
                case(
                    (
                        Contains.child.in_(list(source_info.left_ancestors)),
                        Contains.child,
                    ),
                    (
                        Contains.child.in_(list(source_info.right_ancestors))
                        if source_info.right_ancestors
                        else Contains.child.notin_(list(source_info.left_ancestors)),
                        Contains.child,
                    ),
                )
            ).label("children"),
            func.min(Probabilities.probability).label("probability"),
        )
        .join(valid_pairs, valid_pairs.c.parent == Contains.parent)
        .join(
            Probabilities,
            and_(
                Probabilities.cluster == Contains.parent,
                Probabilities.resolution == resolution.resolution_id,
            ),
        )
        .group_by(Contains.parent)
    ).cte("pairs")

    # Final select to properly split out left and right
    final_select = select(
        pairs.c.id,
        pairs.c.children[1].label("left_id"),
        pairs.c.children[2].label("right_id"),
        pairs.c.probability,
    )

    with MBDB.get_adbc_connection() as conn:
        return sql_to_df(
            stmt=compile_sql(final_select),
            connection=conn,
            return_type="arrow",
        )



"""Common Arrow utilities."""

from io import BytesIO
from typing import Final

import pyarrow as pa
import pyarrow.parquet as pq

SCHEMA_MB_IDS: Final[pa.Schema] = pa.schema(
    [("id", pa.int64()), ("source_pk", pa.large_string())]
)
"""Data transfer schema for Matchbox IDs keyed to primary keys."""

SCHEMA_INDEX: Final[pa.Schema] = pa.schema(
    [("hash", pa.large_binary()), ("source_pk", pa.large_list(pa.large_string()))]
)
"""Data transfer schema for data to be indexed in Matchbox."""

SCHEMA_RESULTS: Final[pa.Schema] = pa.schema(
    [
        ("left_id", pa.uint64()),
        ("right_id", pa.uint64()),
        ("probability", pa.uint8()),
    ]
)
"""Data transfer schema for the results of a deduplication or linking process."""


def table_to_buffer(table: pa.Table) -> BytesIO:
    """Converts an Arrow table to a BytesIO buffer."""
    sink = BytesIO()
    pq.write_table(table, sink)
    sink.seek(0)
    return sink



"""Utilities for querying and matching in the PostgreSQL backend."""

from typing import TypeVar

import pyarrow as pa
from sqlalchemy import BIGINT, Engine, and_, cast, func, literal, null, select, union
from sqlalchemy.orm import Session
from sqlalchemy.sql.selectable import CTE, Select

from matchbox.common.db import sql_to_df
from matchbox.common.exceptions import (
    MatchboxResolutionNotFoundError,
    MatchboxSourceNotFoundError,
)
from matchbox.common.sources import Match, SourceAddress
from matchbox.server.postgresql.db import MBDB
from matchbox.server.postgresql.orm import (
    Clusters,
    ClusterSourcePK,
    Contains,
    Probabilities,
    Resolutions,
    Sources,
)
from matchbox.server.postgresql.utils.db import compile_sql

T = TypeVar("T")


def _get_dataset_source(
    source_name_address: SourceAddress, session: Session
) -> Sources:
    """Converts the named address of source to a Sources ORM object."""
    source = (
        session.query(Sources)
        .filter(
            Sources.full_name == source_name_address.full_name,
            Sources.warehouse_hash == source_name_address.warehouse_hash,
        )
        .first()
    )
    if source is None:
        raise MatchboxSourceNotFoundError(
            address=str(source_name_address),
        )

    return source


def _resolve_thresholds(
    lineage_truths: dict[str, float],
    resolution: Resolutions,
    threshold: int | None,
) -> dict[int, float]:
    """Resolves final thresholds for each resolution in the lineage based on user input.

    Args:
        lineage_truths: Dict from with resolution hash -> cached truth
        resolution: The target resolution being used for clustering
        threshold: User-supplied threshold value

    Returns:
        Dict mapping resolution hash to their final threshold values
    """
    resolved_thresholds = {}

    for resolution_id, default_truth in lineage_truths.items():
        # Dataset
        if default_truth is None:
            resolved_thresholds[resolution_id] = None
            continue

        # Model
        if threshold is None:
            resolved_thresholds[resolution_id] = default_truth
        elif isinstance(threshold, int):
            resolved_thresholds[resolution_id] = (
                threshold
                if resolution_id == resolution.resolution_id
                else default_truth
            )
        else:
            raise ValueError(f"Invalid threshold type: {type(threshold)}")

    return resolved_thresholds


def _get_valid_clusters_for_resolution(resolution_id: int, threshold: int) -> Select:
    """Get clusters that meet the threshold for a specific resolution."""
    return select(Probabilities.cluster.label("cluster")).where(
        and_(
            Probabilities.resolution == resolution_id,
            Probabilities.probability >= threshold,
        )
    )


def _union_valid_clusters(lineage_thresholds: dict[int, float]) -> Select:
    """Creates a CTE of clusters that are valid for any resolution in the lineage.

    Each resolution may have a different threshold.
    """
    valid_clusters = None

    for resolution_id, threshold in lineage_thresholds.items():
        if threshold is None:
            # This is a dataset - get all its clusters through ClusterSourcePK
            resolution_valid = (
                select(ClusterSourcePK.cluster_id.label("cluster"))
                .join(Sources, Sources.source_id == ClusterSourcePK.source_id)
                .where(Sources.resolution_id == resolution_id)
                .distinct()
            )
        else:
            # This is a model - get clusters meeting threshold
            resolution_valid = _get_valid_clusters_for_resolution(
                resolution_id, threshold
            )

        if valid_clusters is None:
            valid_clusters = resolution_valid
        else:
            valid_clusters = union(valid_clusters, resolution_valid)

    if valid_clusters is None:
        # Handle empty lineage case
        return select(cast(null(), BIGINT).label("cluster")).where(False)

    return valid_clusters.cte("valid_clusters").prefix_with("MATERIALIZED")


def _build_valid_contains(valid_clusters_cte: CTE, name: str) -> CTE:
    """Filters the Contains table to only include relationships with valid parents."""
    valid_contains = select(Contains.child, Contains.parent).where(
        Contains.parent.in_(select(valid_clusters_cte.c.cluster))
    )

    return valid_contains.cte(name).prefix_with("MATERIALIZED")


def _resolve_cluster_hierarchy(
    dataset_source: Sources,
    truth_resolution: Resolutions,
    engine: Engine,
    threshold: int | None = None,
) -> Select:
    """Resolves the final cluster assignments for all records in a dataset.

    Args:
        dataset_source: Source object of the dataset to query
        truth_resolution: Resolution object representing the point of truth
        engine: Engine for database connection
        threshold: Optional threshold value

    Returns:
        SQLAlchemy Select statement that will resolve to (hash, id) pairs, where
        hash is the ultimate parent cluster hash and id is the original record ID
    """
    with Session(engine) as session:
        dataset_resolution = session.get(Resolutions, dataset_source.resolution_id)
        if dataset_resolution is None:
            raise MatchboxSourceNotFoundError()

        try:
            lineage_truths = truth_resolution.get_lineage_to_dataset(
                dataset=dataset_resolution
            )
        except ValueError as e:
            raise MatchboxResolutionNotFoundError(
                f"Invalid resolution lineage: {str(e)}"
            ) from e

        thresholds = _resolve_thresholds(
            lineage_truths=lineage_truths,
            resolution=truth_resolution,
            threshold=threshold,
        )

        # Get clusters and contains valid across all resolutions in lineage
        valid_clusters = _union_valid_clusters(thresholds)
        valid_contains = _build_valid_contains(valid_clusters, name="valid_contains")

        # Get base mapping of IDs to clusters
        mapping_base = (
            select(
                Clusters.cluster_id.label("cluster_id"),
                ClusterSourcePK.source_pk.label("source_pk"),
            )
            .join(ClusterSourcePK, ClusterSourcePK.cluster_id == Clusters.cluster_id)
            .where(
                and_(
                    Clusters.cluster_id.in_(select(valid_clusters.c.cluster)),
                    ClusterSourcePK.source_id == dataset_source.source_id,
                )
            )
            .cte("mapping_base")
            .prefix_with("MATERIALIZED")
        )

        # Build recursive hierarchy CTE
        hierarchy = (
            # Base case: direct parents
            select(
                mapping_base.c.cluster_id.label("original_cluster"),
                mapping_base.c.cluster_id.label("child"),
                valid_contains.c.parent.label("parent"),
                literal(1).label("level"),
            )
            .select_from(mapping_base)
            .join(
                valid_contains,
                valid_contains.c.child == mapping_base.c.cluster_id,
                isouter=True,
            )
            .cte("hierarchy", recursive=True)
        )

        # Recursive case
        recursive = (
            select(
                hierarchy.c.original_cluster,
                hierarchy.c.parent.label("child"),
                valid_contains.c.parent.label("parent"),
                (hierarchy.c.level + 1).label("level"),
            )
            .select_from(hierarchy)
            .join(valid_contains, valid_contains.c.child == hierarchy.c.parent)
            .where(hierarchy.c.parent.is_not(None))  # Only recurse on non-leaf nodes
        )

        hierarchy = hierarchy.union_all(recursive)

        # Get highest parents
        highest_parents = (
            select(
                hierarchy.c.original_cluster,
                hierarchy.c.parent.label("highest_parent"),
                hierarchy.c.level,
            )
            .distinct(hierarchy.c.original_cluster)
            .order_by(hierarchy.c.original_cluster, hierarchy.c.level.desc())
            .cte("highest_parents")
        )

        # Final mapping with coalesced results
        final_mapping = (
            select(
                mapping_base.c.source_pk,
                func.coalesce(
                    highest_parents.c.highest_parent, mapping_base.c.cluster_id
                ).label("final_parent"),
            )
            .select_from(mapping_base)
            .join(
                highest_parents,
                highest_parents.c.original_cluster == mapping_base.c.cluster_id,
                isouter=True,
            )
            .cte("final_mapping")
        )

        # Final select statement
        return select(
            final_mapping.c.final_parent.label("id"), final_mapping.c.source_pk
        )


def query(
    source_address: SourceAddress,
    resolution_name: str | None = None,
    threshold: int | None = None,
    limit: int = None,
) -> pa.Table:
    """Queries Matchbox and the Source warehouse to retrieve linked data.

    Takes the dictionaries of tables and fields outputted by selectors and
    queries database for them. If a "point of truth" resolution is supplied, will
    attach the clusters this data belongs to.

    To accomplish this, the function:

    * Iterates through each selector, and
        * Retrieves its data in Matchbox according to the optional point of truth,
        including its hash and cluster hash
        * Retrieves its raw data from its Source's warehouse
        * Joins the two together
    * Unions the results, one row per item of data in the warehouses

    Returns:
        A table containing the requested data from each table, unioned together,
        with the hash key of each row in Matchbox
    """
    engine = MBDB.get_engine()
    with Session(engine) as session:
        dataset_source = _get_dataset_source(source_address, session)
        dataset_resolution = session.get(Resolutions, dataset_source.resolution_id)

        if resolution_name:
            truth_resolution = (
                session.query(Resolutions)
                .filter(Resolutions.name == resolution_name)
                .first()
            )
            if truth_resolution is None:
                raise MatchboxResolutionNotFoundError(resolution_name=resolution_name)
        else:
            truth_resolution = dataset_resolution

        id_query = _resolve_cluster_hierarchy(
            dataset_source=dataset_source,
            truth_resolution=truth_resolution,
            threshold=threshold,
            engine=engine,
        )

        if limit:
            id_query = id_query.limit(limit)

        with MBDB.get_adbc_connection() as conn:
            mb_ids = sql_to_df(
                stmt=compile_sql(id_query),
                connection=conn,
                return_type="arrow",
            )

        return mb_ids


def _build_unnested_clusters() -> CTE:
    """Create CTE that unnests cluster IDs for easier joining."""
    return (
        select(
            Clusters.cluster_id,
            ClusterSourcePK.source_id.label("dataset"),
            ClusterSourcePK.source_pk,
        )
        .select_from(Clusters)
        .join(ClusterSourcePK, ClusterSourcePK.cluster_id == Clusters.cluster_id)
        .cte("unnested_clusters")
        .prefix_with("MATERIALIZED")
    )


def _find_source_cluster(
    unnested_clusters: CTE, source_dataset_id: int, source_pk: str
) -> Select:
    """Find the initial cluster containing the source primary key."""
    return (
        select(unnested_clusters.c.cluster_id)
        .select_from(unnested_clusters)
        .where(
            and_(
                unnested_clusters.c.dataset == source_dataset_id,
                unnested_clusters.c.source_pk == source_pk,
            )
        )
        .scalar_subquery()
    )


def _build_hierarchy_up(
    source_cluster: Select, valid_clusters: CTE | None = None
) -> CTE:
    """Build recursive CTE that finds all parent clusters.

    Args:
        source_cluster: Subquery that finds starting cluster
        valid_clusters: Optional CTE of valid clusters to filter by
    """
    # Pre-filter Contains table if valid_clusters is provided
    contains_table = Contains
    child_col = Contains.child
    parent_col = Contains.parent

    if valid_clusters is not None:
        contains_table = _build_valid_contains(valid_clusters, name="valid_contains_up")
        child_col = contains_table.c.child
        parent_col = contains_table.c.parent

    # Base case: direct parents
    base = (
        select(
            source_cluster.label("original_cluster"),
            source_cluster.label("child"),
            parent_col.label("parent"),
            literal(1).label("level"),
        )
        .select_from(contains_table)
        .where(child_col == source_cluster)
    )

    hierarchy_up = base.cte("hierarchy_up", recursive=True)

    # Recursive case
    recursive = (
        select(
            hierarchy_up.c.original_cluster,
            hierarchy_up.c.parent.label("child"),
            parent_col.label("parent"),
            (hierarchy_up.c.level + 1).label("level"),
        )
        .select_from(hierarchy_up)
        .join(contains_table, child_col == hierarchy_up.c.parent)
    )

    return hierarchy_up.union_all(recursive)


def _find_highest_parent(hierarchy_up: CTE) -> Select:
    """Find the topmost parent cluster from the hierarchy."""
    return (
        select(hierarchy_up.c.parent)
        .order_by(hierarchy_up.c.level.desc())
        .limit(1)
        .scalar_subquery()
    )


def _build_hierarchy_down(
    highest_parent: Select, unnested_clusters: CTE, valid_clusters: CTE | None = None
) -> CTE:
    """Build recursive CTE that finds all child clusters and their IDs.

    Args:
        highest_parent: Subquery that finds top cluster
        unnested_clusters: CTE with unnested cluster IDs
        valid_clusters: Optional CTE of valid clusters to filter by
    """
    # Pre-filter Contains table if valid_clusters is provided
    contains_table = Contains
    child_col = Contains.child
    parent_col = Contains.parent

    if valid_clusters is not None:
        contains_table = _build_valid_contains(
            valid_clusters, name="valid_contains_down"
        )
        child_col = contains_table.c.child
        parent_col = contains_table.c.parent

    # Base case: Get both direct children and their IDs
    base = (
        select(
            highest_parent.label("parent"),
            child_col.label("child"),
            literal(1).label("level"),
            unnested_clusters.c.dataset.label("dataset"),
            unnested_clusters.c.source_pk.label("source_pk"),
        )
        .select_from(contains_table)
        .join_from(
            contains_table,
            unnested_clusters,
            unnested_clusters.c.cluster_id == child_col,
            isouter=True,
        )
        .where(parent_col == highest_parent)
    )

    hierarchy_down = base.cte("hierarchy_down", recursive=True)

    # Recursive case: Get both intermediate nodes AND their leaf records
    recursive = (
        select(
            hierarchy_down.c.parent,
            child_col.label("child"),
            (hierarchy_down.c.level + 1).label("level"),
            unnested_clusters.c.dataset.label("dataset"),
            unnested_clusters.c.source_pk.label("source_pk"),
        )
        .select_from(hierarchy_down)
        .join_from(
            hierarchy_down,
            contains_table,
            parent_col == hierarchy_down.c.child,
        )
        .join_from(
            contains_table,
            unnested_clusters,
            unnested_clusters.c.cluster_id == child_col,
            isouter=True,
        )
        .where(hierarchy_down.c.source_pk.is_(None))  # Only recurse on non-leaf nodes
    )

    return hierarchy_down.union_all(recursive)


def _build_match_query(
    source_pk: str,
    dataset_source: Sources,
    resolution_name: str,
    session: Session,
    threshold: int | None = None,
) -> Select:
    """Builds the SQL query that powers the match function."""
    # Get truth resolution
    truth_resolution = (
        session.query(Resolutions).filter(Resolutions.name == resolution_name).first()
    )
    if truth_resolution is None:
        raise MatchboxResolutionNotFoundError(resolution_name=resolution_name)

    # Get resolution lineage and resolve thresholds
    lineage_truths = truth_resolution.get_lineage()
    thresholds = _resolve_thresholds(
        lineage_truths=lineage_truths,
        resolution=truth_resolution,
        threshold=threshold,
    )

    # Get valid clusters across all resolutions
    valid_clusters = _union_valid_clusters(thresholds)

    # Build the query components
    unnested = _build_unnested_clusters()
    source_cluster = _find_source_cluster(unnested, dataset_source.source_id, source_pk)
    hierarchy_up = _build_hierarchy_up(source_cluster, valid_clusters)
    highest = _find_highest_parent(hierarchy_up)
    hierarchy_down = _build_hierarchy_down(highest, unnested, valid_clusters)

    # Get all matched IDs
    final_stmt = (
        select(
            hierarchy_down.c.parent.label("cluster"),
            hierarchy_down.c.dataset,
            hierarchy_down.c.source_pk,
        )
        .distinct()
        .select_from(hierarchy_down)
    )

    return final_stmt


def match(
    engine: Engine,
    source_pk: str,
    source: SourceAddress,
    targets: list[SourceAddress],
    resolution_name: str,
    threshold: int | None = None,
) -> list[Match]:
    """Matches an ID in a source dataset and returns the keys in the targets.

    To accomplish this, the function:

    * Reconstructs the resolution lineage from the specified resolution
    * Iterates through each target, and
        * Retrieves its cluster hash according to the resolution
        * Retrieves all other IDs in the cluster in the source dataset
        * Retrieves all other IDs in the cluster in the target dataset
    * Returns the results as Match objects, one per target
    """
    with Session(engine) as session:
        # Get all matches for source_pk in all possible targets
        dataset_source = _get_dataset_source(source, session)

        match_stmt = _build_match_query(
            source_pk=source_pk,
            dataset_source=dataset_source,
            resolution_name=resolution_name,
            session=session,
            threshold=threshold,
        )

        matches = session.execute(match_stmt).all()

        # Return matches in target sources only
        cluster = None
        matches_by_source_id: dict[int, set] = {}
        for cluster_id, source_id, id_in_source in matches:
            if cluster is None:
                cluster = cluster_id
            if source_id not in matches_by_source_id:
                matches_by_source_id[source_id] = set()
            matches_by_source_id[source_id].add(id_in_source)

        result = []
        for target_address in targets:
            target_source = _get_dataset_source(target_address, session)
            match_obj = Match(
                cluster=cluster,
                source=source,
                source_id=matches_by_source_id.get(dataset_source.source_id, set()),
                target=target_address,
                target_id=matches_by_source_id.get(target_source.source_id, set()),
            )
            result.append(match_obj)

        return result



"""Core functionalities of the Matchbox client."""

from matchbox.client.helpers.cleaner import cleaner, cleaners
from matchbox.client.helpers.comparison import comparison
from matchbox.client.helpers.selector import select

__all__ = (
    # Cleaners
    "cleaner",
    "cleaners",
    # Comparisons
    "comparison",
    # Selectors
    "select",
)



import copy
from typing import Any, Callable

import pandas as pd
import pyarrow as pa
import pytest
from pandas.testing import assert_frame_equal
from sqlalchemy import Engine, Table, create_engine

from matchbox.client.helpers.selector import Match
from matchbox.common.db import fullname_to_prefix
from matchbox.common.exceptions import MatchboxSourceColumnError
from matchbox.common.factories.sources import source_factory, source_from_tuple
from matchbox.common.sources import Source, SourceAddress, SourceColumn


def test_source_address_compose():
    """Correct addresses are generated from engines and table names."""
    pg = create_engine("postgresql://user:fakepass@host:1234/db")  # trufflehog:ignore
    pg_host = create_engine(
        "postgresql://user:fakepass@host2:1234/db"  # trufflehog:ignore
    )
    pg_port = create_engine(
        "postgresql://user:fakepass@host:4321/db"  # trufflehog:ignore
    )
    pg_db = create_engine(
        "postgresql://user:fakepass@host:1234/db2"  # trufflehog:ignore
    )
    pg_user = create_engine(
        "postgresql://user2:fakepass@host:1234/db"  # trufflehog:ignore
    )
    pg_password = create_engine(
        "postgresql://user:fakepass2@host:1234/db"  # trufflehog:ignore
    )
    pg_dialect = create_engine(
        "postgresql+psycopg://user:fakepass@host:1234/db"  # trufflehog:ignore
    )

    sqlite = create_engine("sqlite:///foo.db")
    sqlite_name = create_engine("sqlite:///bar.db")

    different_wh_hashes = set(
        [
            SourceAddress.compose(pg, "tablename").warehouse_hash,
            SourceAddress.compose(pg_host, "tablename").warehouse_hash,
            SourceAddress.compose(pg_port, "tablename").warehouse_hash,
            SourceAddress.compose(pg_db, "tablename").warehouse_hash,
            SourceAddress.compose(sqlite, "tablename").warehouse_hash,
            SourceAddress.compose(sqlite_name, "tablename").warehouse_hash,
        ]
    )
    different_wh_hashes_str = set([str(sa) for sa in different_wh_hashes])

    assert len(different_wh_hashes) == 6
    assert len(different_wh_hashes_str) == 6

    same_wh_hashes = set(
        [
            SourceAddress.compose(pg, "tablename").warehouse_hash,
            SourceAddress.compose(pg_user, "tablename").warehouse_hash,
            SourceAddress.compose(pg_password, "tablename").warehouse_hash,
            SourceAddress.compose(pg_dialect, "tablename").warehouse_hash,
        ]
    )
    same_wh_hashes_str = set([str(sa) for sa in same_wh_hashes])

    assert len(same_wh_hashes) == 1
    assert len(same_wh_hashes_str) == 1

    same_table_name = set(
        [
            SourceAddress.compose(pg, "tablename").full_name,
            SourceAddress.compose(sqlite, "tablename").full_name,
        ]
    )
    same_table_name_str = set([str(sa) for sa in same_table_name])

    assert len(same_table_name) == 1
    assert len(same_table_name_str) == 1


def test_source_address_format_columns():
    """Column names can get a standard prefix from a table name."""
    address1 = SourceAddress(full_name="foo", warehouse_hash=b"bar")
    address2 = SourceAddress(full_name="foo.bar", warehouse_hash=b"bar")

    assert address1.format_column("col") == "foo_col"
    assert address2.format_column("col") == "foo_bar_col"


def test_source_set_engine(sqlite_warehouse: Engine):
    """Engine can be set on Source."""
    source_testkit = source_factory(engine=sqlite_warehouse)

    # We can set engine with correct column specification
    source = source_testkit.source.set_engine(sqlite_warehouse)
    assert isinstance(source, Source)

    # Error is raised with wrong engine
    with pytest.raises(ValueError, match="engine does not match"):
        wrong_engine = create_engine("sqlite:///:memory:")
        source.set_engine(wrong_engine)


def test_source_check_columns(sqlite_warehouse: Engine):
    """Source columns are checked against the warehouse."""
    source_testkit = source_factory(
        features=[{"name": "b", "base_generator": "random_int", "sql_type": "BIGINT"}],
        engine=sqlite_warehouse,
    )
    source_testkit.to_warehouse(engine=sqlite_warehouse)

    # We can set engine with correct column specification
    source = source_testkit.source.set_engine(sqlite_warehouse)
    assert isinstance(source, Source)

    # Error is raised with custom columns
    with pytest.raises(MatchboxSourceColumnError, match="Columns {'c'} not in"):
        source.check_columns(columns=["c"])

    # Error is raised with missing primary key
    new_source = source_testkit.source.model_copy(update={"db_pk": "typo"}).set_engine(
        sqlite_warehouse
    )
    with pytest.raises(
        MatchboxSourceColumnError, match="Primary key typo not available"
    ):
        new_source.check_columns()

    # Error is raised with missing column
    new_source = source_testkit.source.model_copy(
        update={"columns": (SourceColumn(name="c", type="TEXT"),)}
    ).set_engine(sqlite_warehouse)
    with pytest.raises(MatchboxSourceColumnError, match="Column c not available in"):
        new_source.check_columns()

    # Error is raised with wrong type
    new_source = source_testkit.source.model_copy(
        update={"columns": (SourceColumn(name="b", type="TEXT"),)}
    ).set_engine(sqlite_warehouse)
    with pytest.raises(MatchboxSourceColumnError, match="Type BIGINT != TEXT for b"):
        new_source.check_columns()


def test_source_hash_equality(sqlite_warehouse: Engine):
    """__eq__ and __hash__ behave as expected for a Source."""
    # This won't set the engine just yet
    source_testkit = source_factory(engine=sqlite_warehouse)
    source = source_testkit.source
    source_eq = source.model_copy(deep=True)

    source_testkit.to_warehouse(engine=sqlite_warehouse)
    source.set_engine(sqlite_warehouse)

    assert source.engine != source_eq.engine
    assert source == source_eq
    assert hash(source) == hash(source_eq)


def test_source_default_columns(sqlite_warehouse: Engine):
    """Default columns from the warehouse can be assigned to a Source."""
    source_testkit = source_factory(
        features=[
            {"name": "a", "base_generator": "random_int", "sql_type": "BIGINT"},
            {"name": "b", "base_generator": "word", "sql_type": "TEXT"},
        ],
        engine=sqlite_warehouse,
    )

    source_testkit.to_warehouse(engine=sqlite_warehouse)

    expected_columns = (
        SourceColumn(name="a", type="BIGINT"),
        SourceColumn(name="b", type="TEXT"),
    )

    source = source_testkit.source.set_engine(sqlite_warehouse).default_columns()

    assert source.columns == expected_columns
    # We create a new source, but attributes and engine match
    assert source is not source_testkit.source
    assert source == source_testkit.source
    assert source.engine == sqlite_warehouse


def test_source_to_table(sqlite_warehouse: Engine):
    """Convert Source to SQLAlchemy Table."""
    source_testkit = source_factory(engine=sqlite_warehouse)
    source_testkit.to_warehouse(engine=sqlite_warehouse)

    source = source_testkit.source.set_engine(sqlite_warehouse)

    assert isinstance(source.to_table(), Table)


@pytest.mark.parametrize(
    ("converter", "to_pandas_fn"),
    [
        pytest.param(
            lambda src, **kwargs: src.to_pandas(**kwargs),
            lambda df: df,
            id="pandas",
        ),
        pytest.param(
            lambda src, **kwargs: src.to_arrow(**kwargs),
            lambda arrow: arrow.to_pandas(),
            id="arrow",
        ),
        pytest.param(
            lambda src, **kwargs: src.to_polars(**kwargs),
            lambda polars: polars.to_pandas(),
            id="polars",
        ),
    ],
)
def test_source_conversion_methods(
    sqlite_warehouse: Engine,
    converter: Callable[[Any], Any],
    to_pandas_fn: Callable[[Any], Any],
):
    """Check equivalence of Source to Arrow, Pandas or Polars, with options."""
    source_testkit = source_factory(
        features=[
            {"name": "a", "base_generator": "random_int", "sql_type": "BIGINT"},
            {"name": "b", "base_generator": "word", "sql_type": "TEXT"},
        ],
        engine=sqlite_warehouse,
        n_true_entities=2,
    )
    source_testkit.to_warehouse(engine=sqlite_warehouse)
    source = source_testkit.source.set_engine(sqlite_warehouse).default_columns()
    prefix = fullname_to_prefix(source_testkit.source.address.full_name)
    expected_df_prefixed = (
        source_testkit.data.to_pandas().drop(columns=["id"]).add_prefix(prefix)
    )

    # Test basic conversion
    output = converter(source)
    result_df = to_pandas_fn(output)
    assert_frame_equal(
        expected_df_prefixed, result_df, check_like=True, check_dtype=False
    )

    # Test with limit parameter
    output_limited = converter(source, limit=1)
    result_df_limited = to_pandas_fn(output_limited)
    assert_frame_equal(
        expected_df_prefixed.iloc[:1],
        result_df_limited,
        check_like=True,
        check_dtype=False,
    )

    # Test with fields parameter
    output_fields = converter(source, fields=["a"])
    result_df_fields = to_pandas_fn(output_fields)
    assert_frame_equal(
        expected_df_prefixed[[f"{prefix}pk", f"{prefix}a"]],
        result_df_fields,
        check_like=True,
        check_dtype=False,
    )


def test_source_hash_data(sqlite_warehouse: Engine):
    """A Source can output hashed versions of its rows."""
    original = source_factory(
        full_name="original",
        features=[
            {"name": "a", "base_generator": "random_int", "sql_type": "BIGINT"},
            {"name": "b", "base_generator": "word", "sql_type": "TEXT"},
        ],
        engine=sqlite_warehouse,
        n_true_entities=2,
        repetition=1,
    )

    reordered = copy.deepcopy(original)
    reordered.source = original.source.model_copy(
        update={
            "address": original.source.address.model_copy(
                update={"full_name": "reordered"}
            ),
            "columns": (original.source.columns[1], original.source.columns[0]),
        }
    )

    renamed = copy.deepcopy(original)
    renamed.data = renamed.data.rename_columns({"a": "x"})
    renamed.source = original.source.model_copy(
        update={
            "address": original.source.address.model_copy(
                update={"full_name": "renamed"}
            ),
            "columns": (
                original.source.columns[0].model_copy(update={"name": "x"}),
                original.source.columns[1],
            ),
        }
    )

    original.to_warehouse(engine=sqlite_warehouse)
    reordered.to_warehouse(engine=sqlite_warehouse)
    renamed.to_warehouse(engine=sqlite_warehouse)

    original_source = original.source.set_engine(sqlite_warehouse)
    reordered_source = reordered.source.set_engine(sqlite_warehouse)
    renamed_source = renamed.source.set_engine(sqlite_warehouse)

    original_hash = original_source.hash_data(batch_size=3).to_pandas()
    reordered_hash = reordered_source.hash_data().to_pandas()
    renamed_hash = renamed_source.hash_data().to_pandas()

    # Hash have the right shape
    assert len(original_hash) == 2
    assert len(original_hash.source_pk.iloc[0]) == 2
    assert len(original_hash.source_pk.iloc[1]) == 2

    def sort_df(df: pd.DataFrame) -> pd.DataFrame:
        return df.sort_values(by="hash").reset_index(drop=True)

    # Column order does not matter, column names do
    assert sort_df(original_hash).equals(sort_df(reordered_hash))
    assert not sort_df(original_hash).equals(sort_df(renamed_hash))


def test_source_hash_nulls(sqlite_warehouse: Engine):
    """A Source can output hashed versions of rows with nulls."""
    testkit = source_from_tuple(
        data_tuple=({"a": 1.0}, {"a": None}),
        data_pks=["a", "b"],
        full_name="null_test",
        engine=sqlite_warehouse,
    )
    source = testkit.source.set_engine(sqlite_warehouse)
    testkit.to_warehouse(engine=sqlite_warehouse)

    # Test hashing with nulls
    hashed_data = source.hash_data()

    # No nulls in the hash column
    assert pa.compute.count(hashed_data["hash"], mode="only_null").as_py() == 0

    # Test hashing with null PKs
    null_pk_testkit = source_from_tuple(
        data_tuple=({"a": 1}, {"a": 2}, {"a": 3}),
        data_pks=["a", None, None],
        full_name="null_pk_test",
        engine=sqlite_warehouse,
    )

    # Null PKs should error
    with pytest.raises(ValueError):
        source_with_null_pks = null_pk_testkit.source.set_engine(sqlite_warehouse)
        null_pk_testkit.to_warehouse(engine=sqlite_warehouse)
        source_with_null_pks.hash_data()


@pytest.mark.parametrize(
    ("method_name", "return_type"),
    [
        pytest.param("to_arrow", pa.Table, id="to_arrow"),
        pytest.param("to_pandas", pd.DataFrame, id="to_pandas"),
    ],
)
def test_source_data_batching(method_name, return_type, sqlite_warehouse: Engine):
    """Test Source data retrieval methods with batching parameters."""
    # Create a source with multiple rows of data
    source_testkit = source_factory(
        features=[
            {"name": "a", "base_generator": "random_int", "sql_type": "BIGINT"},
            {"name": "b", "base_generator": "word", "sql_type": "TEXT"},
        ],
        engine=sqlite_warehouse,
        n_true_entities=9,
    )
    source_testkit.to_warehouse(engine=sqlite_warehouse)
    source = source_testkit.source.set_engine(sqlite_warehouse).default_columns()

    # Call the method with batching
    method = getattr(source, method_name)
    batch_iterator = method(return_batches=True, batch_size=3)
    batches = list(batch_iterator)

    # Verify we got the expected number of batches
    assert len(batches) == 3
    for batch in batches:
        assert isinstance(batch, return_type)
        assert len(batch) == 3


def test_match_validates():
    """Match objects are validated when they're instantiated."""
    Match(
        cluster=1,
        source=SourceAddress(full_name="test.source", warehouse_hash=b"bar"),
        source_id={"a"},
        target=SourceAddress(full_name="test.target", warehouse_hash=b"bar"),
        target_id={"b"},
    )

    # Missing source_id with target_id
    with pytest.raises(ValueError):
        Match(
            cluster=1,
            source=SourceAddress(full_name="test.source", warehouse_hash=b"bar"),
            target=SourceAddress(full_name="test.target", warehouse_hash=b"bar"),
            target_id={"b"},
        )

    # Missing cluster with target_id
    with pytest.raises(ValueError):
        Match(
            source=SourceAddress(full_name="test.source", warehouse_hash=b"bar"),
            source_id={"a"},
            target=SourceAddress(full_name="test.target", warehouse_hash=b"bar"),
            target_id={"b"},
        )

    # Missing source_id with cluster
    with pytest.raises(ValueError):
        Match(
            cluster=1,
            source=SourceAddress(full_name="test.source", warehouse_hash=b"bar"),
            target=SourceAddress(full_name="test.target", warehouse_hash=b"bar"),
        )



import ast
from functools import partial
from pathlib import Path
from typing import Callable

import duckdb
import pandas as pd
import pyarrow as pa
import pytest

from matchbox.client.clean import drop
from matchbox.client.clean.steps import (
    clean_punctuation,
    expand_abbreviations,
    list_join_to_string,
    remove_stopwords,
    tokenise,
)
from matchbox.client.clean.utils import alias, cleaning_function, unnest_renest

"""
----------------------------
-- Feature cleaning tests --
----------------------------

To avoid bug-prone unit tests for complex cleaning functions like
cmf.clean.company_name, we instead test the constituent
parts, and the methods that build those parts into something complex.

To test company_name, therefore, we test the leaf functions in the stack:

* clean_comp_names
    * clean_punctuation
    * expand_abbreviations
    * tokenise
    * array_except
    * list_join_to_string

And the methods that assemble them:

* cleaning_function
* unnest_renest

See cleaning/ directory for more information on specific tests.

"""


def load_test_data(path: Path) -> tuple[pd.DataFrame, pd.DataFrame]:
    dirty = pd.read_csv(Path(path, "dirty.csv"), converters={"list": ast.literal_eval})
    clean = pd.read_csv(Path(path, "clean.csv"), converters={"list": ast.literal_eval})

    dirty.columns = ["col"]
    clean.columns = ["col"]

    dirty = dirty.convert_dtypes(dtype_backend="pyarrow")
    clean = clean.convert_dtypes(dtype_backend="pyarrow")

    if isinstance(clean["col"][0], list):
        clean["col"] = clean["col"].astype(pd.ArrowDtype(pa.list_(pa.string())))
    if isinstance(dirty["col"][0], list):
        dirty["col"] = dirty["col"].astype(pd.ArrowDtype(pa.list_(pa.string())))

    return dirty, clean


def passthrough(input_column: str) -> str:
    """
    A passthrough cleaning function that does nothing. Helps test more complex
    building functions.
    """
    return f"{input_column}"


remove_stopwords_partial = partial(remove_stopwords, stopwords=["ltd", "plc"])
expand_abbreviations_partial = partial(
    expand_abbreviations, replacements={"co": "company", "ltd": "limited"}
)

cleaning_tests = [
    ("clean_punctuation", clean_punctuation),
    ("remove_stopwords", remove_stopwords_partial),
    ("list_join_to_string", list_join_to_string),
    ("tokenise", tokenise),
    ("expand_abbreviations", expand_abbreviations_partial),
]


@pytest.mark.parametrize("test", cleaning_tests)
def test_basic_functions(test: tuple[str, Callable], test_root_dir: Path):
    """
    Tests whether the basic cleaning functions do what they're supposed
    to. More complex functions should follow from here.
    """
    test_name = test[0]
    test_cleaning_function = test[1]

    dirty, clean = load_test_data(Path(test_root_dir, "client", "cleaning", test_name))

    cleaned = (
        duckdb.sql(
            f"""
        select
            {test_cleaning_function("col")} as col
        from
            dirty
    """
        )
        .arrow()
        .to_pandas(types_mapper=pd.ArrowDtype)
    )

    assert cleaned.equals(clean)


function_tests = [
    ("tokenise", [tokenise]),
    ("pass", [passthrough]),
    (
        "clean_comp_names",
        [
            clean_punctuation,
            expand_abbreviations_partial,
            tokenise,
            remove_stopwords_partial,
            list_join_to_string,
        ],
    ),
]


@pytest.mark.parametrize("test", function_tests)
def test_function(test: tuple[str, Callable], test_root_dir: Path):
    """
    Tests whether the cleaning function is accurately combining basic
    functions.
    """
    test_name = test[0]
    test_cleaning_function = cleaning_function(*test[1])

    dirty, clean = load_test_data(
        Path(test_root_dir, "client", "cleaning", "cleaning_function", test_name)
    )

    cleaned = test_cleaning_function(dirty, column="col")

    assert cleaned.equals(clean)


nest_unnest_tests = [
    ("expand_abbreviations", expand_abbreviations_partial),
    ("pass", passthrough),
]


@pytest.mark.parametrize("test", nest_unnest_tests)
def test_nest_unnest(test: tuple[str, Callable], test_root_dir: Path):
    """
    Tests whether the nest_unnest function is working.
    """
    test_name = test[0]
    test_cleaning_function = cleaning_function(test[1])

    dirty, clean = load_test_data(
        Path(test_root_dir, "client", "cleaning", "unnest_renest", test_name)
    )

    test_cleaning_function_arrayed = unnest_renest(test_cleaning_function)

    cleaned = test_cleaning_function_arrayed(dirty, column="col")

    # Handle arrays being unsortable
    cleaned["col"] = cleaned["col"].astype("string[pyarrow]")
    clean["col"] = clean["col"].astype("string[pyarrow]")

    cleaned = cleaned.sort_values(by="col").reset_index(drop=True)
    clean = clean.sort_values(by="col").reset_index(drop=True)

    assert cleaned.equals(clean)


def test_alias(test_root_dir: Path):
    """
    Tests whether the alias function is working.
    """
    test_cleaning_function = cleaning_function(passthrough)

    dirty, clean = load_test_data(Path(test_root_dir, "client", "cleaning", "alias"))

    alias_function = alias(test_cleaning_function, "foo")

    cleaned = alias_function(dirty, column="col")

    assert "foo" in cleaned.columns


def test_drop(test_root_dir: Path):
    """
    Tests whether the drop function is working.
    """
    dirty, clean = load_test_data(Path(test_root_dir, "client", "cleaning", "alias"))

    cleaned = drop(dirty, column="col")

    assert len(cleaned.columns) == 0



"""Visualisation utilities."""

import rustworkx as rx
from matplotlib.figure import Figure
from rustworkx.visualization import mpl_draw

from matchbox.client._handler import get_resolution_graph
from matchbox.common.graph import ResolutionNodeType


def draw_resolution_graph() -> Figure:
    """Draws the resolution graph."""
    G: rx.PyDiGraph = get_resolution_graph().to_rx()

    node_indices = G.node_indices()
    datasets = {
        G[node_indices[i]]["id"]: i
        for i in node_indices
        if G[node_indices[i]]["type"] == ResolutionNodeType.DATASET.value
    }

    colours = []
    for i in node_indices:
        node_type = G[node_indices[i]]["type"]
        if node_type == ResolutionNodeType.DATASET.value:
            colours.append((0, 0, 1, 0.2))
        elif node_type == ResolutionNodeType.MODEL.value:
            colours.append((1, 0, 0, 0.2))

    return mpl_draw(
        G,
        pos=rx.spring_layout(
            G,
            pos={v: [0, i / 2] for v, i in enumerate(datasets.values())},
            fixed=set(datasets.values()),
        ),
        node_color=colours,
        with_labels=True,
        labels=lambda node: node["name"],
        font_size=8,
    )



"""A deduplication methodology based on a deterministic set of conditions."""

from typing import Type

import duckdb
from pandas import ArrowDtype, DataFrame
from pydantic import Field

from matchbox.client.models.dedupers.base import Deduper, DeduperSettings


class NaiveSettings(DeduperSettings):
    """A data class to enforce the Naive deduper's settings dictionary shape."""

    unique_fields: list[str] = Field(
        description="A list of columns that will form a unique, deduplicated record"
    )


class NaiveDeduper(Deduper):
    """A simple deduper that deduplicates based on a set of boolean conditions."""

    settings: NaiveSettings

    _id_dtype: Type = None

    @classmethod
    def from_settings(cls, id: str, unique_fields: list[str]) -> "NaiveDeduper":
        """Create a NaiveDeduper from a settings dictionary."""
        settings = NaiveSettings(id=id, unique_fields=unique_fields)
        return cls(settings=settings)

    def prepare(self, data: DataFrame) -> None:
        """Prepare the deduper for deduplication."""
        pass

    def dedupe(self, data: DataFrame) -> DataFrame:
        """Deduplicate the dataframe."""
        self._id_dtype = type(data[self.settings.id][0])

        df = data.copy()

        join_clause = []
        for field in self.settings.unique_fields:
            join_clause.append(f"l.{field} = r.{field}")
        join_clause_compiled = " and ".join(join_clause)

        # Generate a new PK to remove row self-match but ALLOW true duplicate
        # rows where all data items are identical in the source
        df["_unique_e4003b"] = range(df.shape[0])

        # We also need to suppress raw.left_id = raw.right_id in cases where
        # we're deduplicating an unnested array of primary keys
        sql = f"""
            select distinct on (list_sort([raw.left_id, raw.right_id]))
                raw.left_id,
                raw.right_id,
                1.0 as probability
            from (
                select
                    l.{self.settings.id} as left_id,
                    r.{self.settings.id} as right_id
                from
                    df l
                inner join df r on
                    (
                        {join_clause_compiled}
                    ) and
                        l._unique_e4003b != r._unique_e4003b
            ) raw
                where raw.left_id != raw.right_id;
        """
        df_arrow = duckdb.sql(sql).arrow()
        res = df_arrow.to_pandas(
            split_blocks=True, self_destruct=True, types_mapper=ArrowDtype
        )
        del df_arrow

        # Convert bytearray back to bytes
        return res.assign(
            left_id=lambda df: df.left_id.apply(self._id_dtype),
            right_id=lambda df: df.right_id.apply(self._id_dtype),
        )



import datetime
from unittest.mock import Mock, call, patch

import pytest
from sqlalchemy import Engine

from matchbox.client.dags import DAG, DedupeStep, IndexStep, LinkStep, StepInput
from matchbox.client.helpers.selector import Selector
from matchbox.client.models.dedupers import NaiveDeduper
from matchbox.client.models.linkers import DeterministicLinker
from matchbox.common.factories.sources import source_factory


def test_step_input_validation(sqlite_warehouse: Engine):
    """Cannot select sources not available to a step."""
    foo = source_factory(full_name="foo", engine=sqlite_warehouse).source
    bar = source_factory(full_name="bar", engine=sqlite_warehouse).source

    i_foo = IndexStep(source=foo)

    d_foo_right = DedupeStep(
        name="d_foo",
        description="",
        left=StepInput(prev_node=i_foo, select={foo: []}),
        model_class=NaiveDeduper,
        settings={},
        truth=1,
    )

    # CASE 1: Reading from Source
    with pytest.raises(ValueError, match="only select"):
        StepInput(prev_node=i_foo, select={bar: []})

    # CASE 2: Reading from previous step
    with pytest.raises(ValueError, match="Cannot select"):
        StepInput(prev_node=d_foo_right, select={bar: []})


def test_model_step_validation(sqlite_warehouse: Engine):
    foo = source_factory(full_name="foo", engine=sqlite_warehouse).source
    bar = source_factory(full_name="bar", engine=sqlite_warehouse).source
    baz = source_factory(full_name="baz", engine=sqlite_warehouse).source

    i_foo = IndexStep(source=foo)
    i_bar = IndexStep(source=bar)
    i_baz = IndexStep(source=baz)

    d_foo = DedupeStep(
        name="d_foo",
        description="",
        left=StepInput(prev_node=i_foo, select={foo: []}),
        model_class=NaiveDeduper,
        settings={},
        truth=1,
    )

    foo_bar = LinkStep(
        name="foo_bar",
        description="",
        left=StepInput(prev_node=d_foo, select={foo: []}, threshold=0.5),
        right=StepInput(prev_node=i_bar, select={bar: []}, threshold=0.7),
        model_class=DeterministicLinker,
        settings={"left_id": "id", "right_id": "id", "comparisons": ""},
        truth=1,
    )

    foo_bar_baz = LinkStep(
        name="foo_bar_baz",
        description="",
        left=StepInput(prev_node=foo_bar, select={foo: [], bar: []}),
        right=StepInput(prev_node=i_baz, select={baz: []}),
        model_class=DeterministicLinker,
        settings={"left_id": "id", "right_id": "id", "comparisons": ""},
        truth=1,
    )

    # Inherit from a source directly
    assert d_foo.sources == {str(foo.address)}

    # Inherit one source from a previous step
    assert foo_bar.sources == {str(foo.address), str(bar.address)}

    # Inherit multiple sources from a previous step
    assert foo_bar_baz.sources == {str(foo.address), str(bar.address), str(baz.address)}


@patch("matchbox.client.dags._handler.index")
def test_index_step_run(handler_index_mock: Mock, sqlite_warehouse: Engine):
    """Tests that an index step correctly calls the index handler."""
    foo = source_factory(full_name="foo", engine=sqlite_warehouse).source.set_engine(
        sqlite_warehouse
    )

    # Test with batch size
    batch_size = 100

    i_foo = IndexStep(source=foo, batch_size=batch_size)
    i_foo.run()

    handler_index_mock.assert_called_once_with(source=foo, batch_size=batch_size)

    # Test without batch size
    handler_index_mock.reset_mock()

    i_foo_no_batch = IndexStep(source=foo, batch_size=None)
    i_foo_no_batch.run()

    handler_index_mock.assert_called_once_with(source=foo, batch_size=None)


@pytest.mark.parametrize(
    "batched",
    [
        pytest.param(False, id="not batched"),
        pytest.param(True, id="batched"),
    ],
)
def test_dedupe_step_run(
    batched: bool,
    sqlite_warehouse: Engine,
):
    """Tests that a dedupe step orchestrates lower-level API correctly."""
    with (
        patch("matchbox.client.dags.make_model") as make_model_mock,
        patch("matchbox.client.dags.process") as process_mock,
        patch("matchbox.client.dags.query") as query_mock,
    ):
        # Complete mock set up
        model_mock = Mock()
        make_model_mock.return_value = model_mock

        results_mock = Mock()
        results_mock.to_matchbox = Mock()

        model_mock.run = Mock(return_value=results_mock)

        # Set up and run deduper
        foo = source_factory(
            full_name="foo", engine=sqlite_warehouse
        ).source.set_engine(sqlite_warehouse)

        i_foo = IndexStep(source=foo)

        d_foo = DedupeStep(
            name="d_foo",
            description="",
            left=StepInput(
                prev_node=i_foo,
                select={foo: []},
                threshold=0.5,
                batch_size=100 if batched else None,
            ),
            model_class=NaiveDeduper,
            settings={"id": "id", "unique_fields": []},
            truth=1,
        )

        d_foo.run()

        # Right data is queried
        query_mock.assert_called_once_with(
            [Selector(engine=sqlite_warehouse, address=foo.address, fields=[])],
            return_type="pandas",
            threshold=d_foo.left.threshold,
            resolution_name=d_foo.left.name,
            only_indexed=True,
            batch_size=100 if batched else None,
            return_batches=False,
        )
        # Data is pre-processed
        process_mock.assert_called_once_with(
            query_mock.return_value, d_foo.left.cleaners
        )

        # Model is created and run
        make_model_mock.assert_called_once_with(
            model_name=d_foo.name,
            description=d_foo.description,
            model_class=d_foo.model_class,
            model_settings=d_foo.settings,
            left_data=process_mock.return_value,
            left_resolution=d_foo.left.name,
        )
        model_mock.run.assert_called_once()

        # Results are stored
        model_mock.run().to_matchbox.assert_called_once()
        assert model_mock.truth == 1


@pytest.mark.parametrize(
    "batched",
    [
        pytest.param(False, id="not batched"),
        pytest.param(True, id="batched"),
    ],
)
def test_link_step_run(
    batched: bool,
    sqlite_warehouse: Engine,
):
    """Tests that a link step orchestrates lower-level API correctly."""
    with (
        patch("matchbox.client.dags.make_model") as make_model_mock,
        patch("matchbox.client.dags.process") as process_mock,
        patch("matchbox.client.dags.query") as query_mock,
    ):
        # Complete mock set up
        model_mock = Mock()
        make_model_mock.return_value = model_mock

        results_mock = Mock()
        results_mock.to_matchbox = Mock()

        model_mock.run = Mock(return_value=results_mock)

        # Set up and run deduper
        foo = source_factory(
            full_name="foo", engine=sqlite_warehouse
        ).source.set_engine(sqlite_warehouse)
        bar = source_factory(
            full_name="bar", engine=sqlite_warehouse
        ).source.set_engine(sqlite_warehouse)

        i_foo = IndexStep(source=foo)
        i_bar = IndexStep(source=bar)

        foo_bar = LinkStep(
            name="foo_bar",
            description="",
            left=StepInput(
                prev_node=i_foo,
                select={foo: []},
                threshold=0.5,
                batch_size=100 if batched else None,
            ),
            right=StepInput(
                prev_node=i_bar,
                select={bar: []},
                threshold=0.7,
                batch_size=100 if batched else None,
            ),
            model_class=DeterministicLinker,
            settings={"left_id": "id", "right_id": "id", "comparisons": ""},
            truth=1,
        )

        foo_bar.run()

        # Right data is queried
        assert query_mock.call_count == 2
        assert query_mock.call_args_list[0] == call(
            [Selector(engine=sqlite_warehouse, address=foo.address, fields=[])],
            return_type="pandas",
            threshold=foo_bar.left.threshold,
            resolution_name=foo_bar.left.name,
            only_indexed=True,
            batch_size=100 if batched else None,
            return_batches=False,
        )
        assert query_mock.call_args_list[1] == call(
            [Selector(engine=sqlite_warehouse, address=bar.address, fields=[])],
            return_type="pandas",
            threshold=foo_bar.right.threshold,
            resolution_name=foo_bar.right.name,
            only_indexed=True,
            batch_size=100 if batched else None,
            return_batches=False,
        )

        # Data is pre-processed
        assert process_mock.call_count == 2
        assert process_mock.call_args_list[0] == call(
            query_mock.return_value, foo_bar.left.cleaners
        )
        assert process_mock.call_args_list[1] == call(
            query_mock.return_value, foo_bar.right.cleaners
        )

        # Model is created and run
        make_model_mock.assert_called_once_with(
            model_name=foo_bar.name,
            description=foo_bar.description,
            model_class=foo_bar.model_class,
            model_settings=foo_bar.settings,
            left_data=process_mock.return_value,
            left_resolution=foo_bar.left.name,
            right_data=process_mock.return_value,
            right_resolution=foo_bar.right.name,
        )
        model_mock.run.assert_called_once()

        # Results are stored
        model_mock.run().to_matchbox.assert_called_once()
        assert model_mock.truth == 1


@patch("matchbox.client.dags._handler.index")
@patch.object(DedupeStep, "run")
@patch.object(LinkStep, "run")
def test_dag_runs(
    link_run: Mock, dedupe_run: Mock, handler_index: Mock, sqlite_warehouse: Engine
):
    """A legal DAG can be built and run."""
    # Assemble DAG
    dag = DAG()

    # Set up constituents
    foo = source_factory(full_name="foo", engine=sqlite_warehouse).source
    bar = source_factory(full_name="bar", engine=sqlite_warehouse).source
    baz = source_factory(full_name="baz", engine=sqlite_warehouse).source

    # Structure: Sources can be added directly, with and without IndexStep
    i_foo = IndexStep(source=foo, batch_size=100)
    dag.add_steps(i_foo)

    i_bar, i_baz = dag.add_sources(bar, baz, batch_size=200)

    assert set(dag.nodes.keys()) == {
        str(foo.address),
        str(bar.address),
        str(baz.address),
    }

    # Structure: IndexSteps can be deduped
    d_foo = DedupeStep(
        name="d_foo",
        description="",
        left=StepInput(prev_node=i_foo, select={foo: []}),
        model_class=NaiveDeduper,
        settings={},
        truth=1,
    )

    # Structure:
    # - IndexSteps can be passed directly to linkers
    # - Or, linkers can take dedupers
    foo_bar = LinkStep(
        left=StepInput(
            prev_node=d_foo,
            select={foo: []},
        ),
        right=StepInput(
            prev_node=i_bar,
            select={bar: []},
        ),
        name="foo_bar",
        description="",
        model_class=DeterministicLinker,
        settings={},
        truth=1,
    )

    # Structure: Linkers can take other linkers
    foo_bar_baz = LinkStep(
        left=StepInput(
            prev_node=foo_bar,
            select={foo: [], bar: []},
            cleaners={},
        ),
        right=StepInput(
            prev_node=i_baz,
            select={baz: []},
            cleaners={},
        ),
        name="foo_bar_baz",
        description="",
        model_class=DeterministicLinker,
        settings={},
        truth=1,
    )

    dag.add_steps(d_foo, foo_bar, foo_bar_baz)
    assert set(dag.nodes.keys()) == {
        str(foo.address),
        str(bar.address),
        str(baz.address),
        d_foo.name,
        foo_bar.name,
        foo_bar_baz.name,
    }

    # Prepare DAG
    dag.prepare()
    s_foo, s_bar, s_d_foo, s_foo_bar, s_foo_bar_baz = (
        dag.sequence.index(str(foo.address)),
        dag.sequence.index(str(bar.address)),
        dag.sequence.index(d_foo.name),
        dag.sequence.index(foo_bar.name),
        dag.sequence.index(foo_bar_baz.name),
    )
    assert s_foo < s_d_foo < s_foo_bar < s_foo_bar_baz
    assert s_bar < s_foo_bar < s_foo_bar_baz

    # Run DAG
    dag.run()

    assert handler_index.call_count == 3

    # Verify sources and batch sizes passed to handler.index
    calls = {
        call.kwargs["source"]: call.kwargs["batch_size"]
        for call in handler_index.call_args_list
    }

    assert calls[foo] == 100
    assert calls[bar] == 200
    assert calls[baz] == 200

    # Verify the right sources were sent to index
    assert {
        handler_index.call_args_list[0].kwargs["source"],
        handler_index.call_args_list[1].kwargs["source"],
        handler_index.call_args_list[2].kwargs["source"],
    } == {foo, bar, baz}

    assert dedupe_run.call_count == 1
    assert link_run.call_count == 2

    # Reset mocks to test the start argument
    handler_index.reset_mock()
    dedupe_run.reset_mock()
    link_run.reset_mock()

    # Run DAG again, starting from foo_bar step
    dag.run(start="foo_bar")

    # Verify only steps from foo_bar onward were executed
    assert handler_index.call_count == 0
    assert dedupe_run.call_count == 0
    assert link_run.call_count == 2

    # Reset mocks to test the finish argument
    handler_index.reset_mock()
    dedupe_run.reset_mock()
    link_run.reset_mock()

    # Run DAG with finish at foo_bar step (should not execute foo_bar_baz)
    dag.run(finish="foo_bar")

    # Verify steps up to foo_bar were executed but foo_bar_baz was not
    assert handler_index.call_count == 3
    assert dedupe_run.call_count == 1
    assert link_run.call_count == 1

    # Reset mocks again to test combining start and finish
    handler_index.reset_mock()
    dedupe_run.reset_mock()
    link_run.reset_mock()

    # Run from d_foo to foo_bar (skipping sources at start and foo_bar_baz at end)
    dag.run(start="d_foo", finish="foo_bar")

    # Verify only the specified segment was executed
    assert handler_index.call_count == 0
    assert dedupe_run.call_count == 1
    assert link_run.call_count == 1


def test_dag_missing_dependency(sqlite_warehouse: Engine):
    """Steps cannot be added before their dependencies."""
    foo = source_factory(full_name="foo", engine=sqlite_warehouse).source

    i_foo = IndexStep(source=foo)

    d_foo = DedupeStep(
        name="d_foo",
        description="",
        left=StepInput(prev_node=i_foo, select={foo: []}),
        model_class=NaiveDeduper,
        settings={},
        truth=1,
    )

    dag = DAG()
    with pytest.raises(ValueError, match="Dependency"):
        dag.add_steps(d_foo)


def test_dag_name_clash(sqlite_warehouse: Engine):
    """Names across sources and steps must be unique."""
    foo = source_factory(full_name="foo", engine=sqlite_warehouse).source
    bar = source_factory(full_name="bar", engine=sqlite_warehouse).source

    i_foo = IndexStep(source=foo)
    i_bar = IndexStep(source=bar)

    d_foo = DedupeStep(
        name="d_foo",
        description="",
        left=StepInput(prev_node=i_foo, select={foo: []}),
        model_class=NaiveDeduper,
        settings={},
        truth=1,
    )
    d_bar_wrong = DedupeStep(
        # Name clash!
        name="d_foo",
        description="",
        left=StepInput(prev_node=i_bar, select={bar: []}),
        model_class=NaiveDeduper,
        settings={},
        truth=1,
    )

    dag = DAG()
    dag.add_steps(i_foo, d_foo)

    with pytest.raises(ValueError, match="already taken"):
        dag.add_steps(d_bar_wrong)

    # DAG is not modified by failed attempt
    assert dag.nodes["d_foo"] == d_foo
    # We didn't overwrite d_foo's dependencies
    assert dag.graph["d_foo"] == [str(foo.address)]


def test_dag_disconnected(sqlite_warehouse: Engine):
    """Nodes cannot be disconnected."""
    foo = source_factory(full_name="foo", engine=sqlite_warehouse).source
    bar = source_factory(full_name="bar", engine=sqlite_warehouse).source

    dag = DAG()
    _ = dag.add_sources(foo, bar)

    with pytest.raises(ValueError, match="disconnected"):
        dag.prepare()


def test_dag_draw(sqlite_warehouse: Engine):
    """Test that the draw method produces a correct string representation of the DAG."""
    # Set up a simple DAG
    dag = DAG()

    foo = source_factory(full_name="foo", engine=sqlite_warehouse).source
    bar = source_factory(full_name="bar", engine=sqlite_warehouse).source
    baz = source_factory(full_name="baz", engine=sqlite_warehouse).source

    i_foo, i_bar, i_baz = dag.add_sources(foo, bar, baz)

    d_foo = DedupeStep(
        name="d_foo",
        description="",
        left=StepInput(prev_node=i_foo, select={foo: []}),
        model_class=NaiveDeduper,
        settings={},
        truth=1,
    )

    foo_bar = LinkStep(
        left=StepInput(prev_node=d_foo, select={foo: []}),
        right=StepInput(prev_node=i_bar, select={bar: []}),
        name="foo_bar",
        description="",
        model_class=DeterministicLinker,
        settings={},
        truth=1,
    )

    foo_bar_baz = LinkStep(
        left=StepInput(prev_node=foo_bar, select={foo: [], bar: []}),
        right=StepInput(prev_node=i_baz, select={baz: []}),
        name="foo_bar_baz",
        description="",
        model_class=DeterministicLinker,
        settings={},
        truth=1,
    )

    dag.add_steps(d_foo, foo_bar, foo_bar_baz)

    # Prepare the DAG and draw it
    dag.prepare()
    tree_str = dag.draw()

    # Test 1: Drawing without timestamps (original behavior)
    tree_str = dag.draw()

    # Verify the structure
    lines = tree_str.strip().split("\n")

    # The root node should be first
    assert lines[0] == "foo_bar_baz"

    # Check that all nodes are present
    node_names = [
        str(foo.address),
        str(bar.address),
        str(baz.address),
        d_foo.name,
        foo_bar.name,
        foo_bar_baz.name,
    ]

    for node in node_names:
        # Either the node name is at the start of a line or after the tree characters
        node_present = any(line.endswith(node) for line in lines)
        assert node_present, f"Node {node} not found in the tree representation"

    # Check that tree has correct formatting with tree characters
    tree_chars = ["└──", "├──", "│"]
    has_tree_chars = any(char in tree_str for char in tree_chars)
    assert has_tree_chars, (
        "Tree representation doesn't use expected formatting characters"
    )

    # Test 2: Drawing with timestamps (status indicators)
    # Set d_foo as processing and foo_bar as completed
    start_time = datetime.datetime.now()
    doing = "d_foo"
    foo_bar.last_run = datetime.datetime.now()

    # Draw the DAG with status indicators
    tree_str_with_status = dag.draw(start_time=start_time, doing=doing)
    status_lines = tree_str_with_status.strip().split("\n")

    # Verify status indicators are present
    status_indicators = ["✅", "🔄", "⏸️"]
    assert any(indicator in tree_str_with_status for indicator in status_indicators)

    # Check specific statuses: foo_bar done, d_foo working, others awaiting
    for line in status_lines:
        if "foo_bar" in line and "foo_bar_baz" not in line:
            assert "✅" in line
        elif "d_foo" in line:
            assert "🔄" in line
        elif any(
            address in line
            for address in [str(foo.address), str(bar.address), str(baz.address)]
        ):
            assert "⏸️" in line

    # Test 3: Check that node names are still present with status indicators
    for node in node_names:
        node_present = any(node in line for line in status_lines)
        assert node_present, (
            f"Node {node} not found in the tree representation with status indicators"
        )

    # Test 4: Drawing with skipped nodes
    skipped_nodes = [str(foo.address), d_foo.name]
    tree_str_with_skipped = dag.draw(
        start_time=start_time, doing=doing, skipped=skipped_nodes
    )
    skipped_lines = tree_str_with_skipped.strip().split("\n")

    # Check that skipped nodes have the skipped indicator
    for line in skipped_lines:
        if any(skipped in line for skipped in skipped_nodes):
            assert "⏭️" in line

    # Test all status indicators together
    doing = "foo_bar_baz"
    tree_str_all_statuses = dag.draw(
        start_time=start_time, doing=doing, skipped=skipped_nodes
    )
    assert all(
        indicator in tree_str_all_statuses for indicator in ["✅", "🔄", "⏸️", "⏭️"]
    )



"""Data transfer objects for Matchbox API."""

from enum import StrEnum
from importlib.metadata import version
from typing import Literal

from pydantic import BaseModel, Field

from matchbox.common.arrow import SCHEMA_INDEX, SCHEMA_RESULTS


class OKMessage(BaseModel):
    """Generic HTTP OK response."""

    status: str = Field(default="OK")
    version: str = Field(default_factory=lambda: version("matchbox-db"))


class BackendCountableType(StrEnum):
    """Enumeration of supported backend countable types."""

    DATASETS = "datasets"
    MODELS = "models"
    DATA = "data"
    CLUSTERS = "clusters"
    CREATES = "creates"
    MERGES = "merges"
    PROPOSES = "proposes"


class ModelResultsType(StrEnum):
    """Enumeration of supported model results types."""

    PROBABILITIES = "probabilities"
    CLUSTERS = "clusters"


class BackendRetrievableType(StrEnum):
    """Enumeration of supported backend retrievable types."""

    SOURCE = "source"
    RESOLUTION = "resolution"


class BackendUploadType(StrEnum):
    """Enumeration of supported backend upload types."""

    INDEX = "index"
    RESULTS = "results"

    @property
    def schema(self):
        """Get the schema for the upload type."""
        return {
            BackendUploadType.INDEX: SCHEMA_INDEX,
            BackendUploadType.RESULTS: SCHEMA_RESULTS,
        }[self]


class ModelType(StrEnum):
    """Enumeration of supported model types."""

    LINKER = "linker"
    DEDUPER = "deduper"


class ModelOperationType(StrEnum):
    """Enumeration of supported model operations."""

    INSERT = "insert"
    UPDATE_TRUTH = "update_truth"
    UPDATE_ANCESTOR_CACHE = "update_ancestor_cache"
    DELETE = "delete"


class ModelMetadata(BaseModel):
    """Metadata for a model."""

    name: str
    description: str
    type: ModelType
    left_resolution: str
    right_resolution: str | None = None  # Only used for linker models


class ModelAncestor(BaseModel):
    """A model's ancestor and its truth value."""

    name: str = Field(..., description="Name of the ancestor model")
    truth: int | None = Field(
        default=None, description="Truth threshold value", ge=0, le=100, strict=True
    )


class ModelOperationStatus(BaseModel):
    """Status response for any model operation."""

    success: bool
    model_name: str
    operation: ModelOperationType
    details: str | None = None

    @classmethod
    def status_409_examples(cls) -> dict:
        """Examples for 409 status code."""
        return {
            "content": {
                "application/json": {
                    "examples": {
                        "confirm_delete": {
                            "summary": "Delete operation requires confirmation. ",
                            "value": cls(
                                success=False,
                                model_name="example_model",
                                operation=ModelOperationType.DELETE,
                                details=(
                                    "This operation will delete the resolutions "
                                    "deduper_1, deduper_2, linker_1, "
                                    "as well as all probabilities they have created. "
                                    "\n\n"
                                    "It won't delete validation associated with these "
                                    "probabilities. \n\n"
                                    "If you're sure you want to continue, rerun with "
                                    "certain=True"
                                ),
                            ).model_dump(),
                        },
                    },
                }
            }
        }

    @classmethod
    def status_500_examples(cls) -> dict:
        """Examples for 500 status code."""
        return {
            "content": {
                "application/json": {
                    "examples": {
                        "unhandled": {
                            "summary": (
                                "Unhandled exception encountered while updating the "
                                "model's truth value."
                            ),
                            "value": cls(
                                success=False,
                                model_name="example_model",
                                operation=ModelOperationType.UPDATE_TRUTH,
                            ).model_dump(),
                        },
                    },
                }
            }
        }


class CountResult(BaseModel):
    """Response model for count results."""

    entities: dict[BackendCountableType, int]


class UploadStatus(BaseModel):
    """Response model for any file upload processes, like Source or Model results."""

    id: str | None = None
    status: Literal[
        "ready", "awaiting_upload", "queued", "processing", "complete", "failed"
    ]
    details: str | None = None
    entity: BackendUploadType | None = None

    _status_code_mapping = {
        "ready": 200,
        "complete": 200,
        "failed": 400,
        "awaiting_upload": 202,
        "queued": 200,
        "processing": 200,
    }

    def get_http_code(self, status: bool) -> int:
        """Get the HTTP status code for the upload status."""
        if self.status == "failed":
            return 400
        return self._status_code_mapping[self.status]

    @classmethod
    def status_400_examples(cls) -> dict:
        """Examples for 400 status code."""
        return {
            "content": {
                "application/json": {
                    "examples": {
                        "expired_id": {
                            "summary": "Upload ID expired",
                            "value": cls(
                                id="example_id",
                                status="failed",
                                details=(
                                    "Upload ID not found or expired. Entries expire "
                                    "after 30 minutes of inactivity, including "
                                    "failed processes."
                                ),
                                entity=BackendUploadType.INDEX,
                            ).model_dump(),
                        },
                        "schema_mismatch": {
                            "summary": "Schema validation error",
                            "value": cls(
                                id="example_id",
                                status="failed",
                                details="Schema mismatch. Expected: ... Got: ...",
                                entity=BackendUploadType.INDEX,
                            ).model_dump(),
                        },
                    },
                }
            }
        }


class NotFoundError(BaseModel):
    """API error for a 404 status code."""

    details: str
    entity: BackendRetrievableType



"""Low-level components of default cleaning functions."""

from matchbox.client.clean.steps.clean_basic import (
    array_except,
    array_intersect,
    clean_punctuation,
    clean_punctuation_except_hyphens,
    dedupe_and_sort,
    expand_abbreviations,
    filter_cdms_number,
    filter_company_number,
    filter_duns_number,
    get_digits_only,
    get_low_freq_char_sig,
    get_postcode_area,
    list_join_to_string,
    periods_to_nothing,
    punctuation_to_spaces,
    regex_extract_list_of_strings,
    regex_remove_list_of_strings,
    remove_notnumbers_leadingzeroes,
    remove_stopwords,
    remove_whitespace,
    to_lower,
    to_upper,
    tokenise,
    trim,
)
from matchbox.client.clean.steps.clean_basic_original import (
    cms_original_clean_cdms_id,
    cms_original_clean_ch_id,
    cms_original_clean_company_name_ch,
    cms_original_clean_company_name_general,
    cms_original_clean_email,
    cms_original_clean_postcode,
)

__all__ = (
    # Basic steps
    "array_except",
    "array_intersect",
    "periods_to_nothing",
    "punctuation_to_spaces",
    "clean_punctuation",
    "clean_punctuation_except_hyphens",
    "dedupe_and_sort",
    "expand_abbreviations",
    "filter_cdms_number",
    "filter_company_number",
    "filter_duns_number",
    "get_digits_only",
    "get_low_freq_char_sig",
    "get_postcode_area",
    "list_join_to_string",
    "regex_extract_list_of_strings",
    "regex_remove_list_of_strings",
    "remove_notnumbers_leadingzeroes",
    "remove_stopwords",
    "remove_whitespace",
    "to_lower",
    "to_upper",
    "tokenise",
    "trim",
    # Original CMS steps
    "cms_original_clean_cdms_id",
    "cms_original_clean_ch_id",
    "cms_original_clean_company_name_ch",
    "cms_original_clean_company_name_general",
    "cms_original_clean_email",
    "cms_original_clean_postcode",
)



from importlib.metadata import version
from typing import Callable
from unittest.mock import Mock, patch

import pyarrow as pa
import pytest
from httpx import Response
from numpy import ndarray
from pandas import DataFrame
from respx import MockRouter
from sqlalchemy import Engine

from matchbox import index, match, process, query
from matchbox.client._handler import create_client
from matchbox.client._settings import ClientSettings
from matchbox.client.clean import company_name, company_number
from matchbox.client.helpers import cleaner, cleaners, comparison, select
from matchbox.client.helpers.selector import Match
from matchbox.common.arrow import SCHEMA_MB_IDS, table_to_buffer
from matchbox.common.dtos import (
    BackendRetrievableType,
    BackendUploadType,
    NotFoundError,
    UploadStatus,
)
from matchbox.common.exceptions import (
    MatchboxResolutionNotFoundError,
    MatchboxServerFileError,
    MatchboxSourceNotFoundError,
)
from matchbox.common.factories.sources import (
    linked_sources_factory,
    source_factory,
    source_from_tuple,
)
from matchbox.common.graph import DEFAULT_RESOLUTION
from matchbox.common.sources import Source, SourceAddress


def test_cleaners():
    cleaner_name = cleaner(function=company_name, arguments={"column": "company_name"})
    cleaner_number = cleaner(
        function=company_number, arguments={"column": "company_number"}
    )
    cleaner_name_number = cleaners(cleaner_name, cleaner_number)

    assert cleaner_name_number is not None


def test_process():
    crn = source_factory().query
    # Duplicate column
    crn = crn.append_column("company_name_again", crn.column("company_name"))

    cleaner_name = cleaner(
        function=company_name,
        arguments={"column": "company_name"},
    )
    # I can add the same function twice
    cleaner_name_again = cleaner(
        function=company_name,
        arguments={"column": "company_name_again"},
    )
    cleaner_number = cleaner(
        function=company_number,
        arguments={"column": "crn"},
    )
    cleaner_name_number = cleaners(cleaner_name, cleaner_name_again, cleaner_number)

    df_name_cleaned = process(data=crn, pipeline=cleaner_name_number)

    assert isinstance(df_name_cleaned, DataFrame)
    assert df_name_cleaned.shape[0] == 10


def test_comparisons():
    comparison_name_id = comparison(
        sql_condition=(
            "l.company_name = r.company_name and l.data_hub_id = r.data_hub_id"
        )
    )

    assert comparison_name_id is not None


def test_create_client():
    mock_settings = ClientSettings(api_root="http://example.com", timeout=20)
    client = create_client(mock_settings)

    assert client.headers.get("X-Matchbox-Client-Version") == version("matchbox_db")
    assert client.base_url == mock_settings.api_root
    assert client.timeout.connect == mock_settings.timeout
    assert client.timeout.pool == mock_settings.timeout
    assert client.timeout.read == 60 * 30
    assert client.timeout.write == 60 * 30


def test_select_default_engine(
    env_setter: Callable[[str, str], None],
    sqlite_warehouse: Engine,
):
    """We can select without explicit engine if default is set."""
    default_engine = sqlite_warehouse.url.render_as_string(hide_password=False)
    env_setter("MB__CLIENT__DEFAULT_WAREHOUSE", default_engine)

    # Select sources
    selection = select("bar")

    # Check selector contains what we expect
    assert selection[0].fields is None
    assert selection[0].address.full_name == "bar"
    assert selection[0].engine.url == sqlite_warehouse.url


def test_select_missing_engine():
    """We must pass an engine if a default is not set"""
    with pytest.raises(ValueError, match="engine"):
        select("test.bar")


def test_select_mixed_style(matchbox_api: MockRouter, sqlite_warehouse: Engine):
    """We can select specific columns from some of the sources"""
    linked = linked_sources_factory(engine=sqlite_warehouse)

    source1 = linked.sources["crn"].source
    source2 = linked.sources["cdms"].source

    selection = select({"crn": ["company_name"]}, "cdms", engine=sqlite_warehouse)

    assert selection[0].fields == ["company_name"]
    assert selection[1].fields is None
    assert selection[0].address == source1.address
    assert selection[1].address == source2.address
    assert selection[0].engine == sqlite_warehouse
    assert selection[1].engine == sqlite_warehouse


def test_query_no_resolution_ok_various_params(
    matchbox_api: MockRouter, sqlite_warehouse: Engine
):
    """Tests that we can avoid passing resolution name, with a variety of parameters."""
    # Dummy data and source
    testkit = source_from_tuple(
        data_tuple=({"a": 1, "b": "2"}, {"a": 10, "b": "20"}),
        data_pks=["0", "1"],
        full_name="foo",
        engine=sqlite_warehouse,
    )
    source = testkit.source.set_engine(sqlite_warehouse)
    address = source.address
    testkit.to_warehouse(engine=sqlite_warehouse)

    # Mock API
    matchbox_api.get(f"/sources/{address.warehouse_hash_b64}/{address.full_name}").mock(
        return_value=Response(200, json=source.model_dump())
    )

    query_route = matchbox_api.get("/query").mock(
        return_value=Response(
            200,
            content=table_to_buffer(
                pa.Table.from_pylist(
                    [
                        {"source_pk": "0", "id": 1},
                        {"source_pk": "1", "id": 2},
                    ],
                    schema=SCHEMA_MB_IDS,
                )
            ).read(),
        )
    )

    selectors = select({"foo": ["a", "b"]}, engine=sqlite_warehouse)

    # Tests with no optional params
    results = query(selectors)
    assert len(results) == 2
    assert {"foo_a", "foo_b", "id"} == set(results.columns)

    assert dict(query_route.calls.last.request.url.params) == {
        "full_name": address.full_name,
        "warehouse_hash_b64": address.warehouse_hash_b64,
    }

    # Tests with optional params
    results = query(selectors, return_type="arrow", threshold=50).to_pandas()
    assert len(results) == 2
    assert {"foo_a", "foo_b", "id"} == set(results.columns)

    assert dict(query_route.calls.last.request.url.params) == {
        "full_name": address.full_name,
        "warehouse_hash_b64": address.warehouse_hash_b64,
        "threshold": "50",
    }


def test_query_multiple_sources(matchbox_api: MockRouter, sqlite_warehouse: Engine):
    """Tests that we can query multiple sources."""
    # Dummy data and source
    testkit1 = source_from_tuple(
        data_tuple=({"a": 1, "b": "2"}, {"a": 10, "b": "20"}),
        data_pks=["0", "1"],
        full_name="foo",
        engine=sqlite_warehouse,
    )
    source1 = testkit1.source.set_engine(sqlite_warehouse)
    address1 = source1.address
    testkit1.to_warehouse(engine=sqlite_warehouse)

    testkit2 = source_from_tuple(
        data_tuple=({"c": "val"}, {"c": "val"}),
        data_pks=["2", "3"],
        full_name="foo2",
        engine=sqlite_warehouse,
    )
    source2 = testkit2.source.set_engine(sqlite_warehouse)
    address2 = source2.address
    testkit2.to_warehouse(engine=sqlite_warehouse)

    # Mock API
    matchbox_api.get(
        f"/sources/{address1.warehouse_hash_b64}/{address1.full_name}"
    ).mock(return_value=Response(200, json=source1.model_dump()))

    matchbox_api.get(
        f"/sources/{address2.warehouse_hash_b64}/{address2.full_name}"
    ).mock(return_value=Response(200, json=source2.model_dump()))

    query_route = matchbox_api.get("/query").mock(
        side_effect=[
            Response(
                200,
                content=table_to_buffer(
                    pa.Table.from_pylist(
                        [
                            {"source_pk": "0", "id": 1},
                            {"source_pk": "1", "id": 2},
                        ],
                        schema=SCHEMA_MB_IDS,
                    )
                ).read(),
            ),
            Response(
                200,
                content=table_to_buffer(
                    pa.Table.from_pylist(
                        [
                            {"source_pk": "2", "id": 1},
                            {"source_pk": "3", "id": 2},
                        ],
                        schema=SCHEMA_MB_IDS,
                    )
                ).read(),
            ),
        ]
        * 2  # 2 calls to `query()` in this test, each querying server twice
    )

    sels = select("foo", {"foo2": ["c"]}, engine=sqlite_warehouse)

    # Validate results
    results = query(sels)
    assert len(results) == 4
    assert {
        # All columns automatically selected for `foo`
        "foo_pk",
        "foo_a",
        "foo_b",
        # Only one column selected for `foo2`
        "foo2_c",
        # The id always comes back
        "id",
    } == set(results.columns)

    assert dict(query_route.calls[-2].request.url.params) == {
        "full_name": address1.full_name,
        "warehouse_hash_b64": address1.warehouse_hash_b64,
        "resolution_name": DEFAULT_RESOLUTION,
    }
    assert dict(query_route.calls[-1].request.url.params) == {
        "full_name": address2.full_name,
        "warehouse_hash_b64": address2.warehouse_hash_b64,
        "resolution_name": DEFAULT_RESOLUTION,
    }

    # It also works with the selectors specified separately
    query([sels[0]], [sels[1]])


@pytest.mark.parametrize(
    "combine_type",
    ["set_agg", "explode"],
)
def test_query_combine_type(
    combine_type: str, matchbox_api: MockRouter, sqlite_warehouse: Engine
):
    """Various ways of combining multiple sources are supported."""
    # Dummy data and source
    testkit1 = source_from_tuple(
        data_tuple=({"col": 20}, {"col": 40}, {"col": 60}),
        data_pks=["0", "1", "2"],
        full_name="foo",
        engine=sqlite_warehouse,
    )
    source1 = testkit1.source.set_engine(sqlite_warehouse)
    address1 = source1.address
    testkit1.to_warehouse(engine=sqlite_warehouse)

    testkit2 = source_from_tuple(
        data_tuple=({"col": "val1"}, {"col": "val2"}, {"col": "val3"}),
        data_pks=["3", "4", "5"],
        full_name="bar",
        engine=sqlite_warehouse,
    )
    source2 = testkit2.source.set_engine(sqlite_warehouse)
    address2 = source2.address
    testkit2.to_warehouse(engine=sqlite_warehouse)

    # Mock API
    matchbox_api.get(
        f"/sources/{address1.warehouse_hash_b64}/{address1.full_name}"
    ).mock(return_value=Response(200, json=source1.model_dump()))

    matchbox_api.get(
        f"/sources/{address2.warehouse_hash_b64}/{address2.full_name}"
    ).mock(return_value=Response(200, json=source2.model_dump()))

    matchbox_api.get("/query").mock(
        side_effect=[
            Response(
                200,
                content=table_to_buffer(
                    pa.Table.from_pylist(
                        [
                            {"source_pk": "0", "id": 1},
                            {"source_pk": "1", "id": 1},
                            {"source_pk": "2", "id": 2},
                        ],
                        schema=SCHEMA_MB_IDS,
                    )
                ).read(),
            ),
            Response(
                200,
                content=table_to_buffer(
                    pa.Table.from_pylist(
                        [
                            # Creating a duplicate value for the same Matchbox ID
                            {"source_pk": "3", "id": 2},
                            {"source_pk": "3", "id": 2},
                            {"source_pk": "4", "id": 3},
                        ],
                        schema=SCHEMA_MB_IDS,
                    )
                ).read(),
            ),
        ]  # two sources to query
    )

    sels = select("foo", "bar", engine=sqlite_warehouse)

    # Validate results
    results = query(sels, combine_type=combine_type)

    if combine_type == "set_agg":
        expected_len = 3
        for _, row in results.drop(columns=["id"]).iterrows():
            for cell in row.values:
                assert isinstance(cell, ndarray)
                # No duplicates
                assert len(cell) == len(set(cell))
    else:
        expected_len = 5

    assert len(results) == expected_len
    assert {
        "foo_pk",
        "foo_col",
        "bar_pk",
        "bar_col",
        "id",
    } == set(results.columns)


def test_query_unindexed_fields(matchbox_api: MockRouter, sqlite_warehouse: Engine):
    """We cannot query unindexed fields when only_indexed=True."""
    # Dummy data and source
    testkit = source_from_tuple(
        data_tuple=({"a": 1, "b": "2"}, {"a": 10, "b": "20"}),
        data_pks=["0", "1"],
        full_name="foo",
        engine=sqlite_warehouse,
    )
    # Drop one column
    source = testkit.source.model_copy(
        update={"columns": (testkit.source.columns[0],)}
    ).set_engine(sqlite_warehouse)
    address = source.address
    testkit.to_warehouse(engine=sqlite_warehouse)

    # Mock API
    matchbox_api.get(f"/sources/{address.warehouse_hash_b64}/{address.full_name}").mock(
        return_value=Response(200, json=source.model_dump())
    )

    matchbox_api.get("/query").mock(
        return_value=Response(
            200,
            content=table_to_buffer(
                pa.Table.from_pylist(
                    [{"source_pk": "0", "id": 1}],
                    schema=SCHEMA_MB_IDS,
                )
            ).read(),
        )
    )

    selectors = select({"foo": ["a", "b"]}, engine=sqlite_warehouse)

    # Verify exception is raised
    with pytest.raises(ValueError, match="unindexed"):
        query(selectors, only_indexed=True)


def test_query_404_resolution(matchbox_api: MockRouter, sqlite_warehouse: Engine):
    # Mock API
    matchbox_api.get("/query").mock(
        return_value=Response(
            404,
            json=NotFoundError(
                details="Resolution 42 not found",
                entity=BackendRetrievableType.RESOLUTION,
            ).model_dump(),
        )
    )

    selectors = select({"foo": ["a", "b"]}, engine=sqlite_warehouse)

    # Test with no optional params
    with pytest.raises(MatchboxResolutionNotFoundError, match="42"):
        query(selectors)


def test_query_404_source_query(matchbox_api: MockRouter, sqlite_warehouse: Engine):
    """Handles source 404 error when querying."""
    # Mock API
    matchbox_api.get("/query").mock(
        return_value=Response(
            404,
            json=NotFoundError(
                details="Source 42 not found",
                entity=BackendRetrievableType.SOURCE,
            ).model_dump(),
        )
    )

    sels = select({"foo": ["a", "b"]}, engine=sqlite_warehouse)

    # Test with no optional params
    with pytest.raises(MatchboxSourceNotFoundError, match="42"):
        query(sels)


def test_query_404_source_get(matchbox_api: MockRouter, sqlite_warehouse: Engine):
    """Handles source 404 error when retrieving source."""
    # Mock API
    address = SourceAddress.compose(full_name="foo", engine=sqlite_warehouse)

    matchbox_api.get(f"/sources/{address.warehouse_hash_b64}/{address.full_name}").mock(
        return_value=Response(
            404,
            json=NotFoundError(
                details="Source 42 not found", entity=BackendRetrievableType.SOURCE
            ).model_dump(),
        )
    )

    matchbox_api.get("/query").mock(
        return_value=Response(
            200,
            content=table_to_buffer(
                pa.Table.from_pylist(
                    [{"source_pk": "0", "id": 1}], schema=SCHEMA_MB_IDS
                )
            ).read(),
        )
    )

    sels = select({"foo": ["a", "b"]}, engine=sqlite_warehouse)

    with pytest.raises(MatchboxSourceNotFoundError, match="42"):
        query(sels)


def test_query_with_batches(matchbox_api: MockRouter, sqlite_warehouse: Engine):
    """Tests that query correctly handles batching options using real warehouse data."""
    # Dummy data and source
    source_testkit = source_from_tuple(
        data_tuple=({"a": 1, "b": "2"}, {"a": 10, "b": "20"}),
        data_pks=["0", "1"],
        full_name="foo",
        engine=sqlite_warehouse,
    )
    source = source_testkit.source.set_engine(sqlite_warehouse)
    address = source.address
    source_testkit.to_warehouse(engine=sqlite_warehouse)

    # Mock API responses
    matchbox_api.get(f"/sources/{address.warehouse_hash_b64}/{address.full_name}").mock(
        return_value=Response(200, json=source.model_dump())
    )
    matchbox_api.get("/query").mock(
        return_value=Response(
            200,
            content=table_to_buffer(
                pa.Table.from_pylist(
                    [
                        {"source_pk": "0", "id": 1},
                        {"source_pk": "1", "id": 2},
                    ],
                    schema=SCHEMA_MB_IDS,
                )
            ).read(),
        )
    )

    sels = select({"foo": ["a", "b"]}, engine=sqlite_warehouse)

    # Test with return_batches=True
    batch_iterator = query(sels, return_batches=True, batch_size=1, return_type="arrow")

    # Check first batch
    first_batch = next(batch_iterator)
    assert isinstance(first_batch, pa.Table)
    assert len(first_batch) == 1
    assert {"foo_a", "foo_b", "id"} == set(first_batch.column_names)

    # Verify we can get the remaining batch
    remaining_batches = list(batch_iterator)
    assert len(remaining_batches) == 1

    # Test with return_batches=False
    results = query(sels, return_batches=False, batch_size=1000, return_type="arrow")

    # Basic verification
    assert isinstance(results, pa.Table)
    assert len(results) == 2
    assert {"foo_a", "foo_b", "id"} == set(results.column_names)


@patch("matchbox.client.helpers.index.Source")
def test_index_success(
    MockSource: Mock, matchbox_api: MockRouter, sqlite_warehouse: Engine
):
    """Test successful indexing flow through the API."""
    # Mock Source
    source = source_factory(
        features=[{"name": "company_name", "base_generator": "company"}],
        engine=sqlite_warehouse,
    )
    mock_source_instance = source.mock
    MockSource.return_value = mock_source_instance

    # Mock the initial source metadata upload
    source_route = matchbox_api.post("/sources").mock(
        return_value=Response(
            202,
            json=UploadStatus(
                id="test-upload-id",
                status="awaiting_upload",
                entity=BackendUploadType.INDEX,
            ).model_dump(),
        )
    )

    # Mock the data upload
    upload_route = matchbox_api.post("/upload/test-upload-id").mock(
        return_value=Response(
            202,
            json=UploadStatus(
                id="test-upload-id", status="complete", entity=BackendUploadType.INDEX
            ).model_dump(),
        )
    )

    # Call the index function
    index(
        full_name=source.source.address.full_name,
        db_pk=source.source.db_pk,
        engine=sqlite_warehouse,
    )

    # Verify the API calls
    source_call = Source.model_validate_json(
        source_route.calls.last.request.content.decode("utf-8")
    )
    assert source_call == source.source
    assert "test-upload-id" in upload_route.calls.last.request.url.path
    assert b"Content-Disposition: form-data;" in upload_route.calls.last.request.content
    assert b"PAR1" in upload_route.calls.last.request.content


@patch("matchbox.client.helpers.index.Source")
@pytest.mark.parametrize(
    "columns",
    [
        pytest.param(["name", "age"], id="string_columns"),
        pytest.param(
            [
                {"name": "name", "type": "TEXT"},
                {"name": "age", "type": "BIGINT"},
            ],
            id="dict_columns",
        ),
        pytest.param(None, id="default_columns"),
    ],
)
def test_index_with_columns(
    MockSource: Mock,
    matchbox_api: MockRouter,
    columns: list[str] | list[dict[str, str]],
    sqlite_warehouse: Engine,
):
    """Test indexing with different column definition formats."""
    # Create source testkit and mock
    source = source_factory(
        features=[
            {"name": "name", "base_generator": "name"},
            {"name": "age", "base_generator": "random_int"},
        ],
        engine=sqlite_warehouse,
    )
    mock_source_instance = source.mock
    MockSource.return_value = mock_source_instance

    # Mock the API endpoints
    source_route = matchbox_api.post("/sources").mock(
        return_value=Response(
            202,
            json=UploadStatus(
                id="test-upload-id",
                status="awaiting_upload",
                entity=BackendUploadType.INDEX,
            ).model_dump(),
        )
    )

    upload_route = matchbox_api.post("/upload/test-upload-id").mock(
        return_value=Response(
            202,
            json=UploadStatus(
                id="test-upload-id", status="complete", entity=BackendUploadType.INDEX
            ).model_dump(),
        )
    )

    # Call index with column definition
    index(
        full_name=source.source.address.full_name,
        db_pk=source.source.db_pk,
        engine=sqlite_warehouse,
        columns=columns,
    )

    # Verify API calls and source creation
    assert source_route.called
    source_call = Source.model_validate_json(
        source_route.calls.last.request.content.decode("utf-8")
    )
    assert source_call == source.source
    assert "test-upload-id" in upload_route.calls.last.request.url.path
    assert b"Content-Disposition: form-data;" in upload_route.calls.last.request.content
    assert b"PAR1" in upload_route.calls.last.request.content
    mock_source_instance.set_engine.assert_called_once_with(sqlite_warehouse)
    if columns:
        mock_source_instance.default_columns.assert_not_called()
    else:
        mock_source_instance.default_columns.assert_called_once()


@patch("matchbox.client.helpers.index.Source")
def test_index_upload_failure(
    MockSource: Mock, matchbox_api: MockRouter, sqlite_warehouse: Engine
):
    """Test handling of upload failures."""
    # Mock Source
    source = source_factory(
        features=[{"name": "company_name", "base_generator": "company"}],
        engine=sqlite_warehouse,
    )
    mock_source_instance = source.mock
    MockSource.return_value = mock_source_instance

    # Mock successful source creation
    source_route = matchbox_api.post("/sources").mock(
        return_value=Response(
            202,
            json=UploadStatus(
                id="test-upload-id",
                status="awaiting_upload",
                entity=BackendUploadType.INDEX,
            ).model_dump(),
        )
    )

    # Mock failed upload
    upload_route = matchbox_api.post("/upload/test-upload-id").mock(
        return_value=Response(
            400,
            json=UploadStatus(
                id="test-upload-id",
                status="failed",
                details="Invalid schema",
                entity=BackendUploadType.INDEX,
            ).model_dump(),
        )
    )

    # Verify the error is propagated
    with pytest.raises(MatchboxServerFileError):
        index(
            full_name=source.source.address.full_name,
            db_pk=source.source.db_pk,
            engine=sqlite_warehouse,
        )

    # Verify API calls
    source_call = Source.model_validate_json(
        source_route.calls.last.request.content.decode("utf-8")
    )
    assert source_call == source.source
    assert "test-upload-id" in upload_route.calls.last.request.url.path
    assert b"Content-Disposition: form-data;" in upload_route.calls.last.request.content
    assert b"PAR1" in upload_route.calls.last.request.content


def test_index_with_batch_size(matchbox_api: MockRouter, sqlite_warehouse: Engine):
    """Test that batch_size is passed correctly to hash_data when indexing."""
    # Dummy data and source
    source_testkit = source_from_tuple(
        data_tuple=({"company_name": "Company A"}, {"company_name": "Company B"}),
        data_pks=["1", "2"],
        full_name="test_companies",
        engine=sqlite_warehouse,
    )
    source = source_testkit.source.set_engine(sqlite_warehouse)
    source_testkit.to_warehouse(engine=sqlite_warehouse)

    # Mock the API endpoints
    source_route = matchbox_api.post("/sources").mock(
        return_value=Response(
            202,
            json=UploadStatus(
                id="test-upload-id",
                status="awaiting_upload",
                entity=BackendUploadType.INDEX,
            ).model_dump(),
        )
    )

    upload_route = matchbox_api.post("/upload/test-upload-id").mock(
        return_value=Response(
            202,
            json=UploadStatus(
                id="test-upload-id", status="complete", entity=BackendUploadType.INDEX
            ).model_dump(),
        )
    )

    # Spy on the hash_data method to verify batch_size
    with patch.object(Source, "hash_data", wraps=source.hash_data) as spy_hash_data:
        # Call index with batch_size
        index(
            full_name=source.address.full_name,
            db_pk=source.db_pk,
            engine=sqlite_warehouse,
            batch_size=1,
        )

        # Verify batch_size was passed to hash_data
        spy_hash_data.assert_called_once()
        assert spy_hash_data.call_args.kwargs["batch_size"] == 1

        # Verify endpoints were called only once, despite multiple batches
        assert source_route.call_count == 1
        assert upload_route.call_count == 1


def test_match_ok(matchbox_api: MockRouter, sqlite_warehouse: Engine):
    """The client can perform the right call for matching."""
    # Set up mocks
    mock_match1 = Match(
        cluster=1,
        source=SourceAddress(full_name="test.source", warehouse_hash=b"bar"),
        source_id={"a"},
        target=SourceAddress(full_name="test.target", warehouse_hash=b"bar"),
        target_id={"b"},
    )
    mock_match2 = Match(
        cluster=1,
        source=SourceAddress(full_name="test.source", warehouse_hash=b"bar"),
        source_id={"a"},
        target=SourceAddress(full_name="test.target2", warehouse_hash=b"bar"),
        target_id={"b"},
    )
    # The standard JSON serialiser does not handle Pydantic objects
    serialised_matches = (
        f"[{mock_match1.model_dump_json()}, {mock_match2.model_dump_json()}]"
    )

    match_route = matchbox_api.get("/match").mock(
        return_value=Response(200, content=serialised_matches)
    )

    # Use match function
    source = select({"test.source": ["a", "b"]}, engine=sqlite_warehouse)
    target1 = select({"test.target1": ["a", "b"]}, engine=sqlite_warehouse)
    target2 = select({"test.target2": ["a", "b"]}, engine=sqlite_warehouse)

    res = match(
        target1,
        target2,
        source=source,
        source_pk="pk1",
        resolution_name="foo",
    )

    # Verify results
    assert len(res) == 2
    assert isinstance(res[0], Match)
    param_set = sorted(match_route.calls.last.request.url.params.multi_items())
    expected_hash_b64 = SourceAddress.compose(
        full_name="irrelevant", engine=sqlite_warehouse
    ).warehouse_hash_b64
    assert param_set == sorted(
        [
            ("target_full_names", "test.target1"),
            ("target_full_names", "test.target2"),
            ("target_warehouse_hashes_b64", expected_hash_b64),
            ("target_warehouse_hashes_b64", expected_hash_b64),
            ("source_full_name", "test.source"),
            ("source_warehouse_hash_b64", expected_hash_b64),
            ("source_pk", "pk1"),
            ("resolution_name", "foo"),
        ]
    )


def test_match_404_resolution(matchbox_api: MockRouter, sqlite_warehouse: Engine):
    """The client can handle a resolution not found error."""
    # Set up mocks
    matchbox_api.get("/match").mock(
        return_value=Response(
            404,
            json=NotFoundError(
                details="Resolution 42 not found",
                entity=BackendRetrievableType.RESOLUTION,
            ).model_dump(),
        )
    )

    # Use match function
    source = select({"test.source": ["a", "b"]}, engine=sqlite_warehouse)
    target = select({"test.target1": ["a", "b"]}, engine=sqlite_warehouse)

    with pytest.raises(MatchboxResolutionNotFoundError, match="42"):
        match(
            target,
            source=source,
            source_pk="pk1",
            resolution_name="foo",
        )


def test_match_404_source(matchbox_api: MockRouter, sqlite_warehouse: Engine):
    """The client can handle a source not found error."""
    # Set up mocks
    matchbox_api.get("/match").mock(
        return_value=Response(
            404,
            json=NotFoundError(
                details="Source 42 not found",
                entity=BackendRetrievableType.SOURCE,
            ).model_dump(),
        )
    )

    # Use match function
    source = select({"test.source": ["a", "b"]}, engine=sqlite_warehouse)
    target = select({"test.target1": ["a", "b"]}, engine=sqlite_warehouse)

    with pytest.raises(MatchboxSourceNotFoundError, match="42"):
        match(
            target,
            source=source,
            source_pk="pk1",
            resolution_name="foo",
        )



"""Functions to transform data between tabular and graph structures."""

import multiprocessing
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
from typing import Callable, Generic, Hashable, Iterable, TypeVar

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import rustworkx as rx

from matchbox.common.hash import hash_values
from matchbox.common.logging import build_progress_bar, logger

T = TypeVar("T", bound=Hashable)


def to_clusters(
    results: pa.Table,
    dtype: pa.DataType = pa.large_binary,
    hash_func: Callable[[*tuple[T, ...]], T] = hash_values,
) -> pa.Table:
    """Converts probabilities into connected components formed at each threshold.

    Args:
        results: Arrow table with columns ['left_id', 'right_id', 'probability']
        dtype: Arrow data type for the parent and child columns
        hash_func: Function to hash the parent and child values

    Returns:
        Arrow table of parent, child, threshold, sorted by probability descending.
    """
    G = rx.PyGraph()
    added: dict[bytes, int] = {}
    components: dict[str, list] = {"parent": [], "child": [], "threshold": []}

    # Sort probabilities descending and select relevant columns
    results = results.sort_by([("probability", "descending")])
    results = results.select(["left_id", "right_id", "probability"])

    # Get unique probability thresholds, sorted
    thresholds = pc.unique(results.column("probability")).sort(order="descending")

    # Process edges grouped by probability threshold
    for prob in thresholds:
        threshold_edges = results.filter(pc.equal(results.column("probability"), prob))

        # Get state before adding this batch of edges
        old_components = {frozenset(comp) for comp in rx.connected_components(G)}

        # Add all nodes and edges at this probability threshold
        edge_values = list(
            zip(
                threshold_edges.column("left_id").to_pylist(),
                threshold_edges.column("right_id").to_pylist(),
                strict=True,
            )
        )
        for left, right in edge_values:
            for hash_val in (left, right):
                if hash_val not in added:
                    idx = G.add_node(hash_val)
                    added[hash_val] = idx

            G.add_edge(added[left], added[right], None)

        new_components = {frozenset(comp) for comp in rx.connected_components(G)}
        changed_components = new_components - old_components

        # For each changed component, add ALL members at current threshold
        for comp in changed_components:
            children = [G.get_node_data(n) for n in comp]
            parent = hash_func(*children)

            components["parent"].extend([parent] * len(children))
            components["child"].extend(children)
            components["threshold"].extend([prob] * len(children))

    return pa.Table.from_pydict(
        components,
        schema=pa.schema(
            [("parent", dtype()), ("child", dtype()), ("threshold", pa.uint8())]
        ),
    )


def graph_results(
    probabilities: pa.Table, all_nodes: Iterable[int] | None = None
) -> tuple[rx.PyDiGraph, np.ndarray, np.ndarray]:
    """Convert probability table to graph representation.

    Args:
        probabilities: PyArrow table with 'left_id', 'right_id' columns
        all_nodes: superset of node identities figuring in probabilities table.
            Used to optionally add isolated nodes to the graph.

    Returns:
        A tuple containing:
        - Rustwork directed graph
        - A list mapping the 'left_id' probabilities to node indices in the graph
        - A list mapping the 'right_id' probabilities to node indices in the graph
    """
    # Create index to use in graph
    unique = pc.unique(
        pa.concat_arrays(
            [
                probabilities["left_id"].combine_chunks(),
                probabilities["right_id"].combine_chunks(),
            ]
        )
    )

    left_indices = pc.index_in(probabilities["left_id"], unique).to_numpy()
    right_indices = pc.index_in(probabilities["right_id"], unique).to_numpy()

    # Create and process graph
    n_nodes = len(unique)
    n_edges = len(probabilities)

    graph = rx.PyGraph(node_count_hint=n_nodes, edge_count_hint=n_edges)
    graph.add_nodes_from(range(n_nodes))

    if all_nodes is not None:
        isolated_nodes = len(set(all_nodes) - set(unique.to_pylist()))
        graph.add_nodes_from(range(isolated_nodes))

    edges = tuple(zip(left_indices, right_indices, strict=True))
    graph.add_edges_from_no_data(edges)

    return graph, left_indices, right_indices


def attach_components_to_probabilities(probabilities: pa.Table) -> pa.Table:
    """Takes an Arrow table of probabilities and adds a component column.

    Expects an Arrow table of column left_id, right_id, probability.

    Returns a table with an additional column, component.
    """
    # Handle empty probabilities
    if len(probabilities) == 0:
        empty_components = pa.array([], type=pa.int64())
        return probabilities.append_column("component", empty_components)

    graph, left_indices, _ = graph_results(probabilities)
    components = rx.connected_components(graph)

    # Convert components to arrays, map back to input to join, and reattach
    component_indices = np.concatenate([np.array(list(c)) for c in components])
    component_labels = np.repeat(
        np.arange(len(components)), [len(c) for c in components]
    )

    node_to_component = np.zeros(graph.num_nodes(), dtype=np.int64)
    node_to_component[component_indices] = component_labels

    edge_components = pa.array(node_to_component[left_indices])

    return probabilities.append_column("component", edge_components).sort_by(
        [("component", "ascending"), ("probability", "descending")]
    )


class DisjointSet(Generic[T]):
    """Disjoint set forest with "path compression" and "union by rank" heuristics.

    This follows implementation from Cormen, Thomas H., et al. Introduction to
    algorithms. MIT press, 2022
    """

    def __init__(self):
        """Initialize the disjoint set."""
        self.parent: dict[T, T] = {}
        self.rank: dict[T, int] = {}

    def _make_set(self, x: T) -> None:
        """Create a new set with a single element x."""
        self.parent[x] = x
        self.rank[x] = 0

    def add(self, x: T) -> None:
        """Add a new element to the disjoint set."""
        if x not in self.parent:
            self._make_set(x)

    def union(self, x: T, y: T) -> None:
        """Merge the sets containing elements x and y."""
        self._link(self._find(x), self._find(y))

    def _link(self, x: T, y: T) -> None:
        """Merge the sets containing elements x and y."""
        if self.rank[x] > self.rank[y]:
            self.parent[y] = x
        else:
            self.parent[x] = y
            if self.rank[x] == self.rank[y]:
                self.rank[y] += 1

    def _find(self, x: T) -> T:
        """Return the representative element of the set containing x."""
        if x not in self.parent:
            self._make_set(x)
            return x

        if x != self.parent[x]:
            self.parent[x] = self._find(self.parent[x])

        return self.parent[x]

    def get_components(self) -> list[set[T]]:
        """Return the connected components of the disjoint set."""
        components = defaultdict(set)
        for x in self.parent:
            root = self._find(x)
            components[root].add(x)
        return list(components.values())


def component_to_hierarchy(
    table: pa.Table,
    dtype: pa.DataType = pa.large_binary,
    hash_func: Callable[[*tuple[T, ...]], T] = hash_values,
) -> pa.Table:
    """Convert pairwise probabilities into a hierarchical representation.

    Assumes data is pre-sorted by probability descending.

    Args:
        table: Arrow Table with columns ['left', 'right', 'probability']
        dtype: Arrow data type for the parent and child columns
        hash_func: Function to hash the parent and child values

    Returns:
        Arrow Table with columns ['parent', 'child', 'probability']
    """
    probs = pc.unique(table["probability"]).sort(order="descending")

    djs = DisjointSet[int]()  # implements connected components
    current_roots: dict[int, set[int]] = defaultdict(set)  # tracks ultimate parents
    hierarchy: list[tuple[int, int, float]] = []  # the output of this function
    seen_components: set[frozenset[int]] = set()  # track previously seen component sets

    table_by_p = defaultdict(list)
    for batch in table.to_batches():
        d = batch.to_pydict()
        for left, right, p in zip(
            d["left_id"], d["right_id"], d["probability"], strict=False
        ):
            table_by_p[p].append([left, right])

    for threshold in probs.to_numpy():
        for left, right in table_by_p[threshold]:
            djs.union(left, right)
            parent = hash_func(left, right)
            hierarchy.extend([(parent, left, threshold), (parent, right, threshold)])
            current_roots[left].add(parent)
            current_roots[right].add(parent)

        for children in djs.get_components():
            if len(children) <= 2:
                continue  # Skip pairs already handled by pairwise probabilities

            # Skip if we've seen this exact component before
            frozen_children = frozenset(children)
            if frozen_children in seen_components:
                continue

            seen_components.add(frozen_children)

            parent = hash_func(*children)
            prev_roots: set[int] = set()
            for child in children:
                prev_roots.update(current_roots[child])
                current_roots[child] = {parent}

            for r in prev_roots:
                hierarchy.append((parent, r, threshold))

    parents, children, probs = zip(*hierarchy, strict=True)
    return pa.table(
        {
            "parent": pa.array(parents, type=dtype()),
            "child": pa.array(children, type=dtype()),
            "probability": pa.array(probs, type=pa.uint8()),
        }
    )


def to_hierarchical_clusters(
    probabilities: pa.Table,
    proc_func: Callable[[pa.Table, pa.DataType], pa.Table] = component_to_hierarchy,
    hash_func: Callable[[*tuple[T, ...]], T] = hash_values,
    dtype: pa.DataType = pa.large_binary,
    parallel: bool = True,
    timeout: int = 300,
    min_rows_per_worker: int = 1000,
) -> pa.Table:
    """Converts a table of pairwise probabilities into a table of hierarchical clusters.

    Args:
        probabilities: Arrow table with columns ['component', 'left', 'right',
            'probability']
        proc_func: Function to process each component
        hash_func: Function to hash the parent and child values
        dtype: Arrow data type for the parent and child columns
        parallel: Whether to work on separate connected components in parallel
        timeout: Maximum seconds to wait for each worker to finish processing.
            Ignored if parallel==False
        min_rows_per_worker: Minimum number of rows to pass to each worker.
            Ignored if parallel==False

    Returns:
        Arrow table with columns ['parent', 'child', 'probability']
    """
    # Handle empty probabilities
    if len(probabilities) == 0:
        return pa.table(
            {
                "parent": pa.array([], type=dtype()),
                "child": pa.array([], type=dtype()),
                "probability": pa.array([], type=pa.uint8()),
            }
        )

    probabilities = probabilities.sort_by([("component", "ascending")])
    components = pc.unique(probabilities["component"])
    n_cores = multiprocessing.cpu_count()
    n_components = len(components)

    logger.info(f"Processing {n_components:,} components using {n_cores} workers")

    # Split table into separate component tables
    component_col = probabilities["component"]
    indices = []
    start_idx = 0

    with build_progress_bar() as progress:
        split_task = progress.add_task(
            "[cyan]Splitting tables...", total=len(component_col)
        )

        for i in range(1, len(component_col)):
            if (component_col[i] != component_col[i - 1]) and (
                (i - start_idx) >= min_rows_per_worker
            ):
                indices.append((start_idx, i))
                start_idx = i
            progress.update(split_task, advance=1)

        indices.append((start_idx, len(component_col)))
        progress.update(split_task, completed=len(component_col))

    component_tables = []
    for start, end in indices:
        idx_array = pa.array(range(start, end))
        component_tables.append(probabilities.take(idx_array))

    # Process components in parallel
    results = []
    with build_progress_bar() as progress:
        process_task = progress.add_task(
            "[green]Processing components...", total=len(component_tables)
        )
        if parallel:
            with ProcessPoolExecutor(max_workers=n_cores) as executor:
                futures = [
                    executor.submit(
                        proc_func, component_table, hash_func=hash_func, dtype=dtype
                    )
                    for component_table in component_tables
                ]
                for future in futures:
                    try:
                        result = future.result(timeout=timeout)
                        results.append(result)
                        progress.update(process_task, advance=1)
                    except TimeoutError:
                        logger.error(
                            f"Component processing timed out after {timeout} seconds"
                        )
                        raise
                    except Exception as e:
                        logger.error(f"Error processing component: {str(e)}")
                        raise
        else:
            results = [
                proc_func(component_table, hash_func=hash_func, dtype=dtype)
                for component_table in component_tables
                if not progress.update(process_task, advance=1)
            ]

    logger.info(f"Completed processing {len(results):,} components successfully")

    # Create empty table if no results
    if not results:
        logger.warning("No results to concatenate")
        return pa.table(
            {
                "parent": pa.array([], type=dtype()),
                "child": pa.array([], type=dtype()),
                "probability": pa.array([], type=pa.uint8()),
            }
        )

    return pa.concat_tables(results)



"""Functions and classes to define, run and register models."""

from typing import Any, ParamSpec, TypeVar

from pandas import DataFrame

from matchbox.client import _handler
from matchbox.client.models.dedupers.base import Deduper
from matchbox.client.models.linkers.base import Linker
from matchbox.client.results import Results
from matchbox.common.dtos import ModelAncestor, ModelMetadata, ModelType
from matchbox.common.exceptions import MatchboxResolutionNotFoundError

P = ParamSpec("P")
R = TypeVar("R")


class Model:
    """Unified model class for both linking and deduping operations."""

    def __init__(
        self,
        metadata: ModelMetadata,
        model_instance: Linker | Deduper,
        left_data: DataFrame,
        right_data: DataFrame | None = None,
    ):
        """Create a new model instance."""
        self.metadata = metadata
        self.model_instance = model_instance
        self.left_data = left_data
        self.right_data = right_data

    def insert_model(self) -> None:
        """Insert the model into the backend database."""
        _handler.insert_model(model=self.metadata)

    @property
    def results(self) -> Results:
        """Retrieve results associated with the model from the database."""
        results = _handler.get_model_results(name=self.metadata.name)
        return Results(probabilities=results, metadata=self.metadata)

    @results.setter
    def results(self, results: Results) -> None:
        """Write results associated with the model to the database."""
        if results.probabilities.shape[0] > 0:
            _handler.add_model_results(
                name=self.metadata.name, results=results.probabilities
            )

    @property
    def truth(self) -> float:
        """Retrieve the truth threshold for the model."""
        truth = _handler.get_model_truth(name=self.metadata.name)
        return _truth_int_to_float(truth)

    @truth.setter
    def truth(self, truth: float) -> None:
        """Set the truth threshold for the model."""
        _handler.set_model_truth(
            name=self.metadata.name, truth=_truth_float_to_int(truth)
        )

    @property
    def ancestors(self) -> dict[str, float]:
        """Retrieve the ancestors of the model."""
        return {
            ancestor.name: _truth_int_to_float(ancestor.truth)
            for ancestor in _handler.get_model_ancestors(name=self.metadata.name)
        }

    @property
    def ancestors_cache(self) -> dict[str, float]:
        """Retrieve the ancestors cache of the model."""
        return {
            ancestor.name: _truth_int_to_float(ancestor.truth)
            for ancestor in _handler.get_model_ancestors_cache(name=self.metadata.name)
        }

    @ancestors_cache.setter
    def ancestors_cache(self, ancestors_cache: dict[str, float]) -> None:
        """Set the ancestors cache of the model."""
        _handler.set_model_ancestors_cache(
            name=self.metadata.name,
            ancestors=[
                ModelAncestor(name=k, truth=_truth_float_to_int(v))
                for k, v in ancestors_cache.items()
            ],
        )

    def delete(self, certain: bool = False) -> bool:
        """Delete the model from the database."""
        result = _handler.delete_model(name=self.metadata.name, certain=certain)
        return result.success

    def run(self) -> Results:
        """Execute the model pipeline and return results."""
        if self.metadata.type == ModelType.LINKER:
            if self.right_data is None:
                raise MatchboxResolutionNotFoundError(
                    "Right dataset required for linking"
                )

            results = self.model_instance.link(
                left=self.left_data, right=self.right_data
            )
        else:
            results = self.model_instance.dedupe(data=self.left_data)

        return Results(
            probabilities=results,
            model=self,
            metadata=self.metadata,
        )


def make_model(
    model_name: str,
    description: str,
    model_class: type[Linker] | type[Deduper],
    model_settings: dict[str, Any],
    left_data: DataFrame,
    left_resolution: str,
    right_data: DataFrame | None = None,
    right_resolution: str | None = None,
) -> Model:
    """Create a unified model instance for either linking or deduping operations.

    Args:
        model_name: Your unique identifier for the model
        description: Description of the model run
        model_class: Either Linker or Deduper class
        model_settings: Configuration settings for the model
        left_data: Primary dataset
        left_resolution: Resolution name for primary model or dataset
        right_data: Secondary dataset (linking only)
        right_resolution: Resolution name for secondary model or dataset (linking only)

    Returns:
        Model: Configured model instance ready for execution
    """
    model_type = (
        ModelType.LINKER if issubclass(model_class, Linker) else ModelType.DEDUPER
    )

    if model_type == ModelType.LINKER and (
        right_data is None or right_resolution is None
    ):
        raise ValueError("Linking requires both right_data and right_resolution")

    model_instance = model_class.from_settings(**model_settings)

    if model_type == ModelType.LINKER:
        model_instance.prepare(left=left_data, right=right_data)
    else:
        model_instance.prepare(data=left_data)

    metadata = ModelMetadata(
        name=model_name,
        description=description,
        type=model_type.value,
        left_resolution=left_resolution,
        right_resolution=right_resolution,
    )

    return Model(
        metadata=metadata,
        model_instance=model_instance,
        left_data=left_data,
        right_data=right_data,
    )


def _truth_float_to_int(truth: float) -> int:
    """Convert user input float truth values to int."""
    if isinstance(truth, float) and 0.0 <= truth <= 1.0:
        return round(truth * 100)
    else:
        raise ValueError(f"Truth value {truth} not a valid probability")


def _truth_int_to_float(truth: int) -> float:
    """Convert backend int truth values to float."""
    if isinstance(truth, int) and 0 <= truth <= 100:
        return float(truth / 100)
    else:
        raise ValueError(f"Truth value {truth} not valid")
